"""This module contains functions that belong to multiple categories. For example,
concat can be used to concat strings, varbinary, arrays, etc.
"""
import functools
from typing import Any, ClassVar, List, Type

import attr

from treeno.base import PrintOptions
from treeno.datatypes import types as type_consts
from treeno.datatypes.builder import array, char, unknown, varbinary, varchar
from treeno.datatypes.conversions import (
    STRING_TYPES,
    common_supertype,
    promote_varchar_to_char,
)
from treeno.expression import (
    OPERATOR_PRECEDENCE,
    Value,
    value_attr,
    wrap_literal_list,
)
from treeno.functions.base import Function, GenericFunction


def _num_array_layers(dtype: type_consts.DataType) -> int:
    if dtype.type_name != type_consts.ARRAY:
        return 0
    return _num_array_layers(dtype.parameters["dtype"]) + 1


def _get_array_concat_type(
    dtypes: List[type_consts.DataType]
) -> type_consts.DataType:
    # We need to find the number of layers of arrays we have, since we can only permit N or N+1 layers and we
    # strip the topmost, e.g.:
    # SELECT TYPEOF(ARRAY[ARRAY[CHAR '3']] || ARRAY['345']);
    #          _col0
    # -----------------------
    #  array(array(char(3)))
    num_layers = [_num_array_layers(dtype) for dtype in dtypes]
    max_layers = max(num_layers)
    assert (
        max_layers - min(num_layers) <= 1
    ), "Incorrect number of array layers in Concatenate"
    stripped_dtypes = [
        dtype.parameters["dtype"] if num_layers == max_layers else dtype
        for dtype, num_layers in zip(dtypes, num_layers)
    ]
    return array(
        dtype=functools.reduce(
            lambda t1, t2: common_supertype(t1, t2), stripped_dtypes
        )
    )


def _get_string_concat_type(
    dtypes: List[type_consts.DataType]
) -> type_consts.DataType:
    char_conversion = any(
        dtype.type_name == type_consts.CHAR for dtype in dtypes
    )
    if not char_conversion:
        return varchar()
    else:
        max_chars = 0
        for dtype in dtypes:
            if dtype.type_name != type_consts.CHAR:
                dtype = promote_varchar_to_char(dtype)
            max_chars += dtype.parameters["max_chars"]
        return char(max_chars=max_chars)


@value_attr
class Concatenate(Function):
    FN_NAME: ClassVar[str] = "CONCAT"
    values: List[Value] = attr.ib(converter=wrap_literal_list)

    def __attrs_post_init__(self) -> None:
        dtypes = [value.data_type for value in self.values]
        type_names = {dtype.type_name for dtype in dtypes}
        # If there's even a single unknown we should give up here.
        if type_consts.UNKNOWN in type_names:
            self.data_type = unknown()
        elif type_consts.ARRAY in type_names:
            self.data_type = _get_array_concat_type(dtypes)
        elif type_names.issubset(STRING_TYPES):
            self.data_type = _get_string_concat_type(dtypes)
        elif type_names == {type_consts.VARBINARY}:
            self.data_type = varbinary()

    @classmethod
    def from_args(
        cls: Type[GenericFunction], *args: Any, **kwargs: Any
    ) -> GenericFunction:
        assert not kwargs, "Keyword arguments not allowed for CONCAT"
        return cls(values=args)

    def sql(self, opts: PrintOptions) -> str:
        return self.to_string(self.values, opts)


# TODO: See expressions.py
OPERATOR_PRECEDENCE[Concatenate] = 5



from abc import ABC
from enum import Enum
from typing import Any, ClassVar, List, Optional

import attr

from treeno.base import PrintMode, PrintOptions, Sql
from treeno.datatypes import types as type_consts
from treeno.datatypes.builder import (
    array,
    bigint,
    boolean,
    double,
    hll,
    map_,
    qdigest,
    tdigest,
    unknown,
    varbinary,
    varchar,
)
from treeno.expression import Star, Value, value_attr, wrap_literal
from treeno.functions.base import FUNCTIONS_TO_NAMES, Function
from treeno.orderby import OrderTerm
from treeno.printer import StatementPrinter, join_stmts, pad
from treeno.util import parenthesize, quote_literal
from treeno.window import NullTreatment, Window


@value_attr
class AggregateFunction(Function, ABC):
    """Aggregate functions are functions that return a single aggregate value per group.
    They have special properties such as the ability to scan over windows using the OVER clause. For example:

    SELECT MAX(a) OVER (PARTITION BY date ORDER BY timestamp ROWS BETWEEN 5 PRECEDING AND CURRENT ROW)
    """

    orderby: Optional[List[OrderTerm]] = attr.ib(default=None, kw_only=True)
    filter_: Optional[Value] = attr.ib(default=None, kw_only=True)
    window: Optional[Window] = attr.ib(default=None, kw_only=True)
    null_treatment: NullTreatment = attr.ib(
        factory=NullTreatment.default, kw_only=True
    )

    def get_constraint_string(self, opts: PrintOptions) -> str:
        constraint_builder = StatementPrinter()
        if self.filter_:
            constraint_builder.add_entry(
                "FILTER", f"(WHERE {self.filter_.sql(opts)})"
            )
        if self.null_treatment != NullTreatment.RESPECT:
            constraint_builder.add_entry(self.null_treatment.value, "NULLS")
        if self.window:
            constraint_builder.add_entry(
                "OVER", parenthesize(self.window.sql(opts))
            )
        return constraint_builder.to_string(opts)

    def get_orderby_string(self, opts: PrintOptions) -> str:
        assert self.orderby, "Orderby must be defined"
        return "ORDER BY " + join_stmts(
            [order.sql(opts) for order in self.orderby], opts
        )

    def to_string(self, values: List[Value], opts: PrintOptions) -> str:
        arg_string = join_stmts([value.sql(opts) for value in values], opts)
        # TODO: We currently pretty print orderby and constraint stringss on the same indentation level.
        # Although sqlstyle.guide doesn't specify what happens when a line gets too long, I think an extra
        # indentation level would be good for visibility.
        if self.orderby:
            arg_string = join_stmts(
                [arg_string, self.get_orderby_string(opts)], opts, delimiter=" "
            )

        call_str = f"{FUNCTIONS_TO_NAMES[type(self)]}({arg_string})"
        constraint_string = self.get_constraint_string(opts)
        if constraint_string:
            spacing = "\n" if opts.mode == PrintMode.PRETTY else " "
            constraint_string = pad(spacing + constraint_string, opts.spaces)
            return call_str + constraint_string
        return call_str


@value_attr
class UnaryAggregateFunction(AggregateFunction, ABC):
    value: Value = attr.ib(converter=wrap_literal)

    def sql(self, opts: PrintOptions) -> str:
        return self.to_string([self.value], opts)


@value_attr
class BinaryStatsFunction(AggregateFunction, ABC):
    y: Value = attr.ib(converter=wrap_literal)
    x: Value = attr.ib(converter=wrap_literal)

    def __attrs_post_init__(self) -> None:
        self.data_type = double()

    def sql(self, opts: PrintOptions) -> str:
        return self.to_string([self.y, self.x], opts)


@value_attr
class Sum(UnaryAggregateFunction):
    FN_NAME: ClassVar[str] = "SUM"

    def __attrs_post_init__(self) -> None:
        self.data_type = self.value.data_type


@value_attr
class Arbitrary(UnaryAggregateFunction):
    FN_NAME: ClassVar[str] = "ARBITRARY"

    def __attrs_post_init__(self) -> None:
        self.data_type = self.value.data_type


@value_attr
class ArrayAgg(UnaryAggregateFunction):
    FN_NAME: ClassVar[str] = "ARRAY_AGG"

    def __attrs_post_init__(self) -> None:
        self.data_type = array(dtype=self.value.data_type)


@value_attr
class Avg(UnaryAggregateFunction):
    FN_NAME: ClassVar[str] = "AVG"

    def __attrs_post_init__(self) -> None:
        self.data_type = double()


@value_attr
class BoolAnd(UnaryAggregateFunction):
    FN_NAME: ClassVar[str] = "BOOL_AND"

    def __attrs_post_init__(self) -> None:
        self.data_type = boolean()


@value_attr
class BoolOr(UnaryAggregateFunction):
    FN_NAME: ClassVar[str] = "BOOL_OR"

    def __attrs_post_init__(self) -> None:
        self.data_type = boolean()


@value_attr
class Checksum(UnaryAggregateFunction):
    FN_NAME: ClassVar[str] = "CHECKSUM"

    def __attrs_post_init__(self) -> None:
        self.data_type = varbinary()


@value_attr
class Count(UnaryAggregateFunction):
    FN_NAME: ClassVar[str] = "COUNT"

    def __attrs_post_init__(self) -> None:
        # TODO: Do we want to include AliasedStar?
        if isinstance(self.value, Star):
            assert (
                not self.orderby
                and self.null_treatment == NullTreatment.default()
            ), "COUNT(*) cannot be used with orderby and null treatment."
        self.data_type = bigint()


@value_attr
class CountIf(UnaryAggregateFunction):
    FN_NAME: ClassVar[str] = "COUNT_IF"

    def __attrs_post_init__(self) -> None:
        self.data_type = bigint()


@value_attr
class Every(UnaryAggregateFunction):
    FN_NAME: ClassVar[str] = "EVERY"

    def __attrs_post_init__(self) -> None:
        self.data_type = boolean()


@value_attr
class GeometricMean(UnaryAggregateFunction):
    FN_NAME: ClassVar[str] = "GEOMETRIC_MEAN"

    def __attrs_post_init__(self) -> None:
        self.data_type = double()


class CountIndication(Enum):
    WITH_COUNT = "WITH COUNT"
    WITHOUT_COUNT = "WITHOUT COUNT"


@attr.s
class OverflowFiller(Sql):
    DEFAULT_FILLER: ClassVar[str] = "..."
    count_indication: CountIndication = attr.ib()
    filler: str = attr.ib(default=DEFAULT_FILLER)

    def sql(self, opts: PrintOptions) -> str:
        return f"TRUNCATE {quote_literal(self.filler)} {self.count_indication.value}"


@value_attr
class ListAgg(AggregateFunction):
    DEFAULT_SEPARATOR: ClassVar[str] = " "
    FN_NAME: ClassVar[str] = "LISTAGG"
    value: Value = attr.ib(converter=wrap_literal)
    separator: str = attr.ib(default=" ")
    # If overflow filler is None, then raise an error. Otherwise truncate
    # with the appropriate filler string.
    overflow_filler: Optional[OverflowFiller] = attr.ib(default=None)

    def __attrs_post_init__(self) -> None:
        assert (
            self.null_treatment == NullTreatment.default()
        ), "LISTAGG currently does not support null treatment"
        assert (
            not self.window
        ), "LISTAGG currently does not support window functions"
        assert not self.filter_, "LISTAGG currently does not support filter"
        assert (
            self.orderby
        ), "LISTAGG requires the WITHIN GROUP(ORDER BY ...) clause"
        self.data_type = varchar()

    def sql(self, opts: PrintOptions) -> str:
        value_string = self.value.sql(opts)
        if self.separator != self.DEFAULT_SEPARATOR:
            value_string = join_stmts(
                [value_string, quote_literal(self.separator)], opts
            )
        # If it's not None, then we could also write ON OVERFLOW ERROR, but it's default so let's omit it.
        if self.overflow_filler is not None:
            value_string = join_stmts(
                [value_string, f"ON OVERFLOW {self.overflow_filler.sql(opts)}"],
                opts,
                delimiter=" ",
            )
        constraint_string = f"WITHIN GROUP ({self.get_orderby_string(opts)})"
        spacing = "\n" if opts.mode == PrintMode.PRETTY else " "
        constraint_string = pad(spacing + constraint_string, opts.spaces)
        return f"{FUNCTIONS_TO_NAMES[type(self)]}({value_string}){constraint_string}"


@value_attr
class Max(AggregateFunction):
    FN_NAME: ClassVar[str] = "MAX"
    value: Value = attr.ib(converter=wrap_literal)
    num_values: Optional[Value] = attr.ib(
        default=None, converter=attr.converters.optional(wrap_literal)
    )

    def __attrs_post_init__(self) -> None:
        self.data_type = self.value.data_type

    def sql(self, opts: PrintOptions) -> str:
        values = [self.value]
        if self.num_values:
            values.append(self.num_values)
        return self.to_string(values, opts)


@value_attr
class MaxBy(AggregateFunction):
    FN_NAME: ClassVar[str] = "MAX_BY"
    value: Value = attr.ib(converter=wrap_literal)
    max_by: Value = attr.ib(converter=wrap_literal)
    num_values: Optional[Value] = attr.ib(
        default=None, converter=attr.converters.optional(wrap_literal)
    )

    def __attrs_post_init__(self) -> None:
        self.data_type = self.value.data_type

    def sql(self, opts: PrintOptions) -> str:
        values = [self.value, self.max_by]
        if self.num_values:
            values.append(self.num_values)
        return self.to_string(values, opts)


@value_attr
class Min(AggregateFunction):
    FN_NAME: ClassVar[str] = "MIN"
    value: Value = attr.ib(converter=wrap_literal)
    num_values: Optional[Value] = attr.ib(
        default=None, converter=attr.converters.optional(wrap_literal)
    )

    def __attrs_post_init__(self) -> None:
        self.data_type = self.value.data_type

    def sql(self, opts: PrintOptions) -> str:
        values = [self.value]
        if self.num_values:
            values.append(self.num_values)
        return self.to_string(values, opts)


@value_attr
class MinBy(AggregateFunction):
    FN_NAME: ClassVar[str] = "MIN_BY"
    value: Value = attr.ib(converter=wrap_literal)
    max_by: Value = attr.ib(converter=wrap_literal)
    num_values: Optional[Value] = attr.ib(
        default=None, converter=attr.converters.optional(wrap_literal)
    )

    def __attrs_post_init__(self) -> None:
        self.data_type = self.value.data_type

    def sql(self, opts: PrintOptions) -> str:
        values = [self.value, self.max_by]
        if self.num_values:
            values.append(self.num_values)
        return self.to_string(values, opts)


@value_attr
class BitwiseAndAgg(UnaryAggregateFunction):
    FN_NAME: ClassVar[str] = "BITWISE_AND_AGG"

    def __attrs_post_init__(self) -> None:
        self.data_type = bigint()


@value_attr
class BitwiseOrAgg(UnaryAggregateFunction):
    FN_NAME: ClassVar[str] = "BITWISE_OR_AGG"

    def __attrs_post_init__(self) -> None:
        self.data_type = bigint()


@value_attr
class Histogram(UnaryAggregateFunction):
    FN_NAME: ClassVar[str] = "HISTOGRAM"

    def __attrs_post_init__(self) -> None:
        self.data_type = map_(
            from_dtype=self.value.data_type, to_dtype=bigint()
        )


@value_attr
class MapAgg(AggregateFunction):
    FN_NAME: ClassVar[str] = "MAP_AGG"
    key: Value = attr.ib(converter=wrap_literal)
    value: Value = attr.ib(converter=wrap_literal)

    def __attrs_post_init__(self) -> None:
        self.data_type = map_(
            from_dtype=self.key.data_type, to_dtype=self.value.data_type
        )

    def sql(self, opts: PrintOptions) -> str:
        return self.to_string([self.key, self.value], opts)


@value_attr
class MapUnion(UnaryAggregateFunction):
    FN_NAME: ClassVar[str] = "MAP_UNION"

    def __attrs_post_init__(self) -> None:
        self.data_type = self.value.data_type


@value_attr
class MultiMapAgg(AggregateFunction):
    FN_NAME: ClassVar[str] = "MULTIMAP_AGG"
    key: Value = attr.ib(converter=wrap_literal)
    value: Value = attr.ib(converter=wrap_literal)

    def __attrs_post_init__(self) -> None:
        self.data_type = map_(
            from_dtype=self.key.data_type,
            to_dtype=array(dtype=self.value.data_type),
        )

    def sql(self, opts: PrintOptions) -> str:
        return self.to_string([self.key, self.value], opts)


@value_attr
class ApproxDistinct(AggregateFunction):
    FN_NAME: ClassVar[str] = "APPROX_DISTINCT"
    key: Value = attr.ib(converter=wrap_literal)
    epsilon: Optional[Value] = attr.ib(
        default=None, converter=attr.converters.optional(wrap_literal)
    )

    def __attrs_post_init__(self) -> None:
        self.data_type = bigint()

    def sql(self, opts: PrintOptions) -> str:
        values = [self.key]
        if self.epsilon:
            values.append(self.epsilon)
        return self.to_string(values, opts)


@value_attr
class ApproxMostFrequent(AggregateFunction):
    FN_NAME: ClassVar[str] = "APPROX_MOST_FREQUENT"
    buckets: Value = attr.ib(converter=wrap_literal)
    value: Value = attr.ib(converter=wrap_literal)
    capacity: Value = attr.ib(converter=wrap_literal)

    def __attrs_post_init__(self) -> None:
        self.data_type = map_(
            from_dtype=self.value.data_type, to_dtype=bigint()
        )

    def sql(self, opts: PrintOptions) -> str:
        return self.to_string([self.buckets, self.value, self.capacity], opts)


@value_attr(init=False)
class ApproxPercentile(AggregateFunction):
    FN_NAME: ClassVar[str] = "APPROX_PERCENTILE"
    value: Value = attr.ib(converter=wrap_literal)
    percentage: Value = attr.ib(converter=wrap_literal)
    weight: Optional[Value] = attr.ib(
        default=None, converter=attr.converters.optional(wrap_literal)
    )

    def __init__(self, value: Value, *args: Any, **kwargs: Any):
        # This function is strange in that the overloads causes positional arguments to mean different things.
        # We have to swap weight and percentage if they're both specified.
        if len(args) == 2:
            args = (args[1], args[0])
        self.__attrs_init__(value, *args, **kwargs)  # type: ignore

        # If percentage is an array, then we can be sure the output type is array
        # otherwise if it's a scalar not-unknown type, then we know the output type is scalar.
        # If we just don't know about the underlying data type then this should return unknown.
        if self.percentage.data_type.type_name == type_consts.ARRAY:
            self.data_type = array(dtype=self.value.data_type)
        elif self.percentage.data_type != unknown():
            self.data_type = self.value.data_type
        else:
            self.data_type = unknown()

    def sql(self, opts: PrintOptions) -> str:
        values = [self.value, self.percentage]
        if self.weight:
            values.append(self.weight)
        return self.to_string(values, opts)


@value_attr
class ApproxSet(UnaryAggregateFunction):
    FN_NAME: ClassVar[str] = "APPROX_SET"

    def __attrs_post_init__(self) -> None:
        self.data_type = hll()


@value_attr
class Merge(UnaryAggregateFunction):
    FN_NAME: ClassVar[str] = "MERGE"

    def __attrs_post_init__(self) -> None:
        if (
            self.value.data_type.type_name == type_consts.TDIGEST
            or self.value.data_type.type_name == type_consts.QDIGEST
        ):
            self.data_type = self.value.data_type
        else:
            self.data_type = hll()


@value_attr
class NumericHistogram(AggregateFunction):
    FN_NAME: ClassVar[str] = "NUMERIC_HISTOGRAM"
    buckets: Value = attr.ib(converter=wrap_literal)
    value: Value = attr.ib(converter=wrap_literal)
    weight: Optional[Value] = attr.ib(
        default=None, converter=attr.converters.optional(wrap_literal)
    )

    def __attrs_post_init__(self) -> None:
        self.data_type = map_(from_dtype=double(), to_dtype=double())

    def sql(self, opts: PrintOptions) -> str:
        values = [self.buckets, self.value]
        if self.weight:
            values.append(self.weight)
        return self.to_string(values, opts)


@value_attr
class QDigestAgg(AggregateFunction):
    FN_NAME: ClassVar[str] = "QDIGEST_AGG"
    value: Value = attr.ib(converter=wrap_literal)
    weight: Optional[Value] = attr.ib(
        default=None, converter=attr.converters.optional(wrap_literal)
    )
    accuracy: Optional[Value] = attr.ib(
        default=None, converter=attr.converters.optional(wrap_literal)
    )

    def __attrs_post_init__(self) -> None:
        self.data_type = qdigest(dtype=self.value.data_type)

    def sql(self, opts: PrintOptions) -> str:
        values = [self.value]
        if self.weight:
            values.append(self.weight)
        if self.accuracy:
            values.append(self.accuracy)
        return self.to_string(values, opts)


@value_attr
class TDigestAgg(AggregateFunction):
    FN_NAME: ClassVar[str] = "TDIGEST_AGG"
    value: Value = attr.ib(converter=wrap_literal)
    weight: Optional[Value] = attr.ib(
        default=None, converter=attr.converters.optional(wrap_literal)
    )

    def __attrs_post_init__(self) -> None:
        self.data_type = tdigest()

    def sql(self, opts: PrintOptions) -> str:
        values = [self.value]
        if self.weight:
            values.append(self.weight)
        return self.to_string(values, opts)


@value_attr
class Corr(BinaryStatsFunction):
    FN_NAME: ClassVar[str] = "CORR"


@value_attr
class CovarPop(BinaryStatsFunction):
    FN_NAME: ClassVar[str] = "COVAR_POP"


@value_attr
class CovarSamp(BinaryStatsFunction):
    FN_NAME: ClassVar[str] = "COVAR_SAMP"


@value_attr
class Kurtosis(BinaryStatsFunction):
    FN_NAME: ClassVar[str] = "KURTOSIS"


@value_attr
class RegrIntercept(BinaryStatsFunction):
    FN_NAME: ClassVar[str] = "REGR_INTERCEPT"


@value_attr
class RegrSlope(BinaryStatsFunction):
    FN_NAME: ClassVar[str] = "REGR_SLOPE"


@value_attr
class Skewness(UnaryAggregateFunction):
    FN_NAME: ClassVar[str] = "SKEWNESS"


@value_attr
class StdDev(UnaryAggregateFunction):
    FN_NAME: ClassVar[str] = "STDDEV"


@value_attr
class StdDevPop(UnaryAggregateFunction):
    FN_NAME: ClassVar[str] = "STDDEV_POP"


@value_attr
class StdDevSamp(UnaryAggregateFunction):
    FN_NAME: ClassVar[str] = "STDDEV_SAMP"


@value_attr
class Variance(UnaryAggregateFunction):
    FN_NAME: ClassVar[str] = "VARIANCE"


@value_attr
class VarPop(UnaryAggregateFunction):
    FN_NAME: ClassVar[str] = "VAR_POP"


@value_attr
class VarSamp(UnaryAggregateFunction):
    FN_NAME: ClassVar[str] = "VAR_SAMP"


# TODO: Add reduce()



from io import StringIO

import nltk
from antlr4 import CommonTokenStream
from antlr4.InputStream import InputStream
from antlr4.ParserRuleContext import ParserRuleContext
from antlr4.tree.Trees import Trees

try:
    from treeno.grammar.gen.SqlBaseLexer import SqlBaseLexer  # type: ignore
    from treeno.grammar.gen.SqlBaseParser import SqlBaseParser  # type: ignore
except ImportError:
    print(
        "SqlBaseLexer and SqlBaseParser not found. Did you run `python setup.py install` or `python setup.py "
        "develop` first?"
    )
    raise


class AST:
    """Basic building block for SQL parse tree. This is not meant to be used directly, but is rather an intermediate
    layer to operate on to create a pypika interface."""

    def __init__(self, sql):
        self._sql = sql

    def parser(self) -> SqlBaseParser:
        lexer = SqlBaseLexer(InputStream(data=self._sql))
        stream = CommonTokenStream(lexer)
        return SqlBaseParser(stream)

    @property
    def text(self):
        return self._sql

    def query(self):
        return self.parser().query()

    def expression(self):
        return self.parser().standaloneExpression()

    def type(self):
        return self.parser().standaloneType()


def tree(ast: AST, node: ParserRuleContext) -> str:
    parenthetical_tree = nltk.Tree.fromstring(
        Trees.toStringTree(node, None, ast.parser())
    )
    sio = StringIO()
    parenthetical_tree.pretty_print(stream=sio)
    return sio.getvalue()



from abc import ABC
from typing import Any, ClassVar, Optional

import attr

from treeno.base import PrintOptions
from treeno.datatypes import types as type_consts
from treeno.datatypes.builder import (
    bigint,
    date,
    double,
    interval,
    time,
    timestamp,
    varchar,
)
from treeno.expression import Value, value_attr, wrap_literal
from treeno.functions.base import FUNCTIONS_TO_NAMES, Function, UnaryFunction
from treeno.util import parenthesize

# I tried even setting hive.timestamp_precision=NANOSECONDS in session properties
# and CURRENT_TIMESTAMP still gives the default precision
DEFAULT_DATETIME_PRECISION: int = 3


@value_attr
class CurrentDate(Function):
    FN_NAME: ClassVar[str] = "CURRENT_DATE"

    def __attrs_post_init__(self) -> None:
        self.data_type = date()

    def sql(self, opts: PrintOptions) -> str:
        return self.FN_NAME


@value_attr
class ParametrizedDateTimeFunction(Function, ABC):
    """This class doesn't inherit from Function because it doesn't have to have parentheses when there's no arguments.
    """

    precision: int = attr.ib(default=DEFAULT_DATETIME_PRECISION)

    def sql(self, opts: PrintOptions) -> str:
        precision_str = ""
        if self.precision != DEFAULT_DATETIME_PRECISION:
            precision_str = parenthesize(str(self.precision))
        return f"{FUNCTIONS_TO_NAMES[type(self)]}{precision_str}"


@value_attr
class CurrentTime(ParametrizedDateTimeFunction):
    FN_NAME: ClassVar[str] = "CURRENT_TIME"

    def __attrs_post_init__(self) -> None:
        self.data_type = time(precision=self.precision)


@value_attr
class CurrentTimestamp(ParametrizedDateTimeFunction):
    FN_NAME: ClassVar[str] = "CURRENT_TIMESTAMP"

    def __attrs_post_init__(self) -> None:
        self.data_type = timestamp(precision=self.precision)


@value_attr
class LocalTime(ParametrizedDateTimeFunction):
    FN_NAME: ClassVar[str] = "LOCALTIME"

    def __attrs_post_init__(self) -> None:
        self.data_type = time(precision=self.precision)


@value_attr
class LocalTimestamp(ParametrizedDateTimeFunction):
    FN_NAME: ClassVar[str] = "LOCALTIMESTAMP"

    def __attrs_post_init__(self) -> None:
        self.data_type = timestamp(precision=self.precision)


@value_attr
class Date(UnaryFunction):
    FN_NAME: ClassVar[str] = "DATE"

    def __attrs_post_init__(self) -> None:
        self.data_type = date()


@value_attr
class LastDayOfMonth(UnaryFunction):
    FN_NAME: ClassVar[str] = "LAST_DAY_OF_MONTH"

    def __attrs_post_init__(self) -> None:
        self.data_type = date()


@value_attr
class FromISO8601Timestamp(UnaryFunction):
    FN_NAME: ClassVar[str] = "FROM_ISO8601_TIMESTAMP"

    def __attrs_post_init__(self) -> None:
        self.data_type = timestamp(precision=3, timezone=True)


@value_attr
class FromISO8601TimestampNanos(UnaryFunction):
    FN_NAME: ClassVar[str] = "FROM_ISO8601_TIMESTAMP_NANOS"

    def __attrs_post_init__(self) -> None:
        self.data_type = timestamp(precision=9, timezone=True)


@value_attr
class FromISO8601Date(UnaryFunction):
    FN_NAME: ClassVar[str] = "FROM_ISO8601_DATE"

    def __attrs_post_init__(self) -> None:
        self.data_type = date()


@value_attr
class AtTimezone(Function):
    FN_NAME: ClassVar[str] = "AT_TIMEZONE"
    value: Value = attr.ib(converter=wrap_literal)
    zone: Value = attr.ib(converter=wrap_literal)

    def __attrs_post_init__(self) -> None:
        # If it's something else like unknown, then we just keep the data type unknown
        dtype = self.value.data_type
        if dtype.type_name == type_consts.TIMESTAMP:
            self.data_type = timestamp(
                precision=dtype.parameters["precision"], timezone=True
            )

    def sql(self, opts: PrintOptions) -> str:
        return self.to_string([self.value, self.zone], opts)


@value_attr
class WithTimezone(Function):
    FN_NAME: ClassVar[str] = "WITH_TIMEZONE"
    value: Value = attr.ib(converter=wrap_literal)
    zone: Value = attr.ib(converter=wrap_literal)

    def __attrs_post_init__(self) -> None:
        dtype = self.value.data_type
        if dtype.type_name == type_consts.TIMESTAMP:
            self.data_type = timestamp(
                precision=dtype.parameters["precision"], timezone=True
            )

    def sql(self, opts: PrintOptions) -> str:
        return self.to_string([self.value, self.zone], opts)


@value_attr
class FromUnixtime(Function):
    FN_NAME: ClassVar[str] = "FROM_UNIXTIME"
    value: Value = attr.ib(converter=wrap_literal)
    zone: Optional[Value] = attr.ib(
        default=None, converter=attr.converters.optional(wrap_literal)
    )
    hours: Optional[Value] = attr.ib(
        default=None, converter=attr.converters.optional(wrap_literal)
    )
    minutes: Optional[Value] = attr.ib(
        default=None, converter=attr.converters.optional(wrap_literal)
    )

    @classmethod
    def from_args(cls, *args: Any, **kwargs: Any) -> "FromUnixtime":
        assert (
            not kwargs
        ), "No keyword arguments allowed for FromUnixTime.from_args"
        if len(args) == 2:
            return FromUnixtime(args[0], zone=args[1])
        if len(args) == 3:
            return FromUnixtime(args[0], hours=args[1], minutes=args[2])
        raise ValueError(f"Unrecognized arguments {args} and {kwargs}")

    def __attrs_post_init__(self) -> None:
        dtype = self.value.data_type
        if dtype.type_name == type_consts.TIMESTAMP:
            self.data_type = timestamp(
                precision=dtype.parameters["precision"], timezone=True
            )

        if self.zone:
            assert (
                not self.hours and not self.minutes
            ), "If zone is specified, can't specify hours or minutes"
        if self.hours or self.minutes:
            assert (
                self.hours and self.minutes and not self.zone
            ), "Hours and minutes must be both specified without zone"

    def sql(self, opts: PrintOptions) -> str:
        values = [self.value]
        if self.zone:
            values.append(self.zone)
        if self.hours:
            assert self.minutes is not None
            values += [self.hours, self.minutes]
        return self.to_string(values, opts)


@value_attr
class FromUnixtimeNanos(UnaryFunction):
    FN_NAME: ClassVar[str] = "FROM_UNIXTIME_NANOS"

    def __attrs_post_init__(self) -> None:
        self.data_type = timestamp(precision=9, timezone=True)


@value_attr
class Now(Function):
    FN_NAME: ClassVar[str] = "NOW"

    def __attrs_post_init__(self) -> None:
        self.data_type = timestamp(precision=3, timezone=True)

    def sql(self, opts: PrintOptions) -> str:
        return self.to_string([], opts)


@value_attr
class ToISO8601(UnaryFunction):
    FN_NAME: ClassVar[str] = "TO_ISO8601"

    def __attrs_post_init__(self) -> None:
        self.data_type = varchar()


@value_attr
class ToMilliseconds(UnaryFunction):
    FN_NAME: ClassVar[str] = "TO_MILLISECONDS"

    def __attrs_post_init__(self) -> None:
        self.data_type = bigint()


@value_attr
class ToUnixtime(UnaryFunction):
    FN_NAME: ClassVar[str] = "TO_UNIXTIME"

    def __attrs_post_init__(self) -> None:
        self.data_type = double()


@value_attr
class DateTrunc(Function):
    FN_NAME: ClassVar[str] = "DATE_TRUNC"
    unit: Value = attr.ib(converter=wrap_literal)
    value: Value = attr.ib(converter=wrap_literal)

    def __attrs_post_init__(self) -> None:
        self.data_type = self.value.data_type

    def sql(self, opts: PrintOptions) -> str:
        return self.to_string([self.unit, self.value], opts)


@value_attr
class DateAdd(Function):
    FN_NAME: ClassVar[str] = "DATE_ADD"
    unit: Value = attr.ib(converter=wrap_literal)
    value: Value = attr.ib(converter=wrap_literal)
    timestamp: Value = attr.ib(converter=wrap_literal)

    def __attrs_post_init__(self) -> None:
        self.data_type = self.timestamp.data_type

    def sql(self, opts: PrintOptions) -> str:
        return self.to_string([self.unit, self.value, self.timestamp], opts)


@value_attr
class DateDiff(Function):
    FN_NAME: ClassVar[str] = "DATE_DIFF"
    unit: Value = attr.ib(converter=wrap_literal)
    timestamp1: Value = attr.ib(converter=wrap_literal)
    timestamp2: Value = attr.ib(converter=wrap_literal)

    def __attrs_post_init__(self) -> None:
        self.data_type = bigint()

    def sql(self, opts: PrintOptions) -> str:
        return self.to_string(
            [self.unit, self.timestamp1, self.timestamp2], opts
        )


@value_attr
class ParseDuration(UnaryFunction):
    FN_NAME: ClassVar[str] = "PARSE_DURATION"

    def __attrs_post_init__(self) -> None:
        self.data_type = interval(from_interval="DAY", to_interval="SECOND")


@value_attr
class HumanReadableSeconds(UnaryFunction):
    FN_NAME: ClassVar[str] = "HUMAN_READABLE_SECONDS"

    def __attrs_post_init__(self) -> None:
        self.data_type = varchar()


# TODO: This isn't the exhaustive list of datetime functions yet.



"""Module for inferring types from python inputs.

Should not be used in datatypes.types to prevent circular imports
"""
from decimal import Decimal
from typing import Any

from treeno.datatypes.builder import (
    bigint,
    boolean,
    char,
    decimal,
    double,
    integer,
    varchar,
)
from treeno.datatypes.types import DataType


def infer_type(value: Any) -> DataType:
    """Infers a trino type from the given python value.
    """
    if isinstance(value, bool):
        return boolean()
    if isinstance(value, int):
        return infer_integral(value)
    if isinstance(value, float):
        return double()
    if isinstance(value, Decimal):
        # NOTE: This is best-effort inference, since the precision of
        # a decimal is encoded in its string value
        return infer_decimal(value)
    if isinstance(value, str):
        # NOTE: We opt for varchar type with no length limit here because the max char limit can be any
        # integer greater than len(value)
        return varchar(max_chars=len(value))
    raise NotImplementedError(
        f"Value {value} with type {type(value).__name__} can't be inferred"
    )


def infer_char(value: str) -> DataType:
    return char(max_chars=len(value))


def infer_integral(value: int) -> DataType:
    return integer() if -(0x7FFFFFFF + 1) <= value <= 0x7FFFFFFF else bigint()


def infer_decimal(decimal_value: Decimal) -> DataType:
    """We must infer a decimal from str and not float because we'll never be sure what the
    actual scale is due to floats being imprecise.
    """
    str_val = str(decimal_value)
    if "." not in str_val:
        return decimal(precision=len(str_val), scale=0)
    else:
        assert str_val.count(".") == 1, f"Malformed decimal str_val {str_val}"
        decimal_point_pos = str_val.index(".")
        return decimal(
            precision=len(str_val) - 1,
            scale=len(str_val) - decimal_point_pos - 1,
        )


def infer_timelike_precision(value: str) -> int:
    # TODO: Use a datetime util for this at some point. Currently datetime in python only supports microseconds :(
    # Number of digits past the dot is its precision
    if "." not in value:
        return 0
    # In case we have timezone
    return len(value.rsplit(".")[-1].split(" ")[0])



from typing import ClassVar

from treeno.base import PrintOptions
from treeno.datatypes.builder import varchar
from treeno.expression import value_attr
from treeno.functions.base import Function


@value_attr
class CurrentUser(Function):
    FN_NAME: ClassVar[str] = "CURRENT_USER"

    def __attrs_post_init__(self) -> None:
        self.data_type = varchar()

    def sql(self, opts: PrintOptions) -> str:
        return self.FN_NAME


@value_attr
class CurrentCatalog(Function):
    FN_NAME: ClassVar[str] = "CURRENT_CATALOG"

    def __attrs_post_init__(self) -> None:
        self.data_type = varchar()

    def sql(self, opts: PrintOptions) -> str:
        return self.FN_NAME


@value_attr
class CurrentSchema(Function):
    FN_NAME: ClassVar[str] = "CURRENT_SCHEMA"

    def __attrs_post_init__(self) -> None:
        self.data_type = varchar()

    def sql(self, opts: PrintOptions) -> str:
        return self.FN_NAME


@value_attr
class CurrentPath(Function):
    FN_NAME: ClassVar[str] = "CURRENT_PATH"

    def __attrs_post_init__(self) -> None:
        self.data_type = varchar()

    def sql(self, opts: PrintOptions) -> str:
        return self.FN_NAME



"""
Converts from our grammar into a buildable query tree.
"""
from decimal import Decimal
from typing import Any, Dict, List, Optional, Tuple, Union

from overrides import overrides

from treeno.datatypes.builder import (
    array,
    boolean,
    double,
    integer,
    interval,
    row,
    unknown,
    varchar,
)
from treeno.datatypes.inference import infer_decimal, infer_integral
from treeno.datatypes.types import FIELDS, TIME, TIMESTAMP, DataType
from treeno.expression import (
    AliasedStar,
    AliasedValue,
    And,
    Array,
    Between,
    Case,
    Cast,
    DistinctFrom,
    Else,
    Field,
    InList,
    InQuery,
    Interval,
    IsNull,
    Lambda,
    Like,
    Literal,
    Or,
    RowConstructor,
    Star,
    Subscript,
    TryCast,
    TypeConstructor,
    Value,
    When,
)
from treeno.functions.aggregate import (
    AggregateFunction,
    CountIndication,
    ListAgg,
    OverflowFiller,
)
from treeno.functions.base import NAMES_TO_FUNCTIONS, Function
from treeno.functions.common import Concatenate
from treeno.functions.session import (
    CurrentCatalog,
    CurrentPath,
    CurrentSchema,
    CurrentUser,
)
from treeno.grammar.gen.SqlBaseParser import SqlBaseParser
from treeno.grammar.gen.SqlBaseVisitor import SqlBaseVisitor
from treeno.grammar.parse import AST
from treeno.groupby import Cube, GroupBy, GroupingSet, GroupingSetList, Rollup
from treeno.orderby import NullOrder, OrderTerm, OrderType
from treeno.relation import (
    AliasedRelation,
    ExceptQuery,
    IntersectQuery,
    Join,
    JoinConfig,
    JoinCriteria,
    JoinOnCriteria,
    JoinType,
    JoinUsingCriteria,
    Lateral,
    Query,
    Relation,
    SampleType,
    SelectQuery,
    SetQuantifier,
    SetQuery,
    Table,
    TableQuery,
    TableSample,
    UnionQuery,
    Unnest,
    ValuesQuery,
)
from treeno.util import nth
from treeno.window import (
    BoundedFrameBound,
    BoundType,
    CurrentFrameBound,
    FrameType,
    NullTreatment,
    UnboundedFrameBound,
    Window,
)


def query_from_sql(sql: str) -> Query:
    return ConvertVisitor().visitQuery(AST(sql).query())


def expression_from_sql(sql: str) -> Value:
    return ConvertVisitor().visitStandaloneExpression(AST(sql).expression())


def type_from_sql(sql: str) -> DataType:
    return ConvertVisitor().visitStandaloneType(AST(sql).type())


def apply_operator(operator: str, *args: Value) -> Value:
    if operator == "AND":
        return args[0] & args[1]
    elif operator == "OR":
        return args[0] | args[1]
    elif operator == "=":
        return args[0] == args[1]
    elif operator == ">":
        return args[0] > args[1]
    elif operator == ">=":
        return args[0] >= args[1]
    elif operator == "<":
        return args[0] < args[1]
    elif operator == "<=":
        return args[0] <= args[1]
    elif operator == "<>":
        return args[0] != args[1]
    elif operator == "+":
        if len(args) == 1:
            return +args[0]
        else:
            assert len(args) == 2
            return args[0] + args[1]
    elif operator == "-":
        if len(args) == 1:
            return -args[0]
        else:
            assert len(args) == 2
            return args[0] - args[1]
    elif operator == "*":
        return args[0] * args[1]
    elif operator == "/":
        return args[0] / args[1]
    elif operator == "%":
        return args[0] % args[1]
    else:
        raise NotImplementedError(f"Unrecognized token {operator}")


def table_from_qualifiers(qualifiers: List[str]) -> Table:
    qualifiers = list(reversed(qualifiers))
    name: str = qualifiers[0]
    schema: Optional[str] = nth(qualifiers, 1)
    catalog: Optional[str] = nth(qualifiers, 2)

    return Table(name=name, schema=schema, catalog=catalog)


class ConvertVisitor(SqlBaseVisitor):
    """Converts the tree into a builder tree in python
    """

    @overrides
    def visitSingleStatement(
        self, ctx: SqlBaseParser.SingleStatementContext
    ) -> Query:
        stmt = ctx.statement()
        if not isinstance(stmt, SqlBaseParser.StatementDefaultContext):
            raise NotImplementedError(
                "Only standard selects are implemented for now"
            )
        return self.visit(ctx.statement())

    @overrides
    def visitStandaloneExpression(
        self, ctx: SqlBaseParser.StandaloneExpressionContext
    ) -> Value:
        return self.visit(ctx.expression())

    @overrides
    def visitStandaloneType(
        self, ctx: SqlBaseParser.StandaloneTypeContext
    ) -> DataType:
        return self.visit(ctx.type_())

    @overrides
    def visitStatementDefault(self, ctx: SqlBaseParser.StatementDefaultContext):
        return self.visit(ctx.query())

    @overrides
    def visitQuery(self, ctx: SqlBaseParser.QueryContext) -> Query:
        query = self.visit(ctx.queryNoWith())
        with_ = ctx.with_()
        if with_:
            named_queries = self.visit(with_)
            query.with_ = named_queries
        return query

    @overrides
    def visitWith_(
        self, ctx: SqlBaseParser.With_Context
    ) -> List[AliasedRelation]:
        assert (
            not ctx.RECURSIVE()
        ), "Recursive with queries currently not supported"
        return [self.visit(named_query) for named_query in ctx.namedQuery()]

    @overrides
    def visitNamedQuery(
        self, ctx: SqlBaseParser.NamedQueryContext
    ) -> AliasedRelation:
        column_aliases = None
        if ctx.columnAliases():
            column_aliases = self.visit(ctx.columnAliases())
        return AliasedRelation(
            relation=self.visit(ctx.query()),
            alias=self.visit(ctx.name),
            column_aliases=column_aliases,
        )

    @overrides
    def visitQueryNoWith(self, ctx: SqlBaseParser.QueryNoWithContext) -> Query:
        query = self.visit(ctx.queryTerm())
        if ctx.ORDER() and ctx.BY():
            query.orderby = [self.visit(item) for item in ctx.sortItem()]
        if ctx.offset:
            query.offset = self.visit(ctx.offset)
        limit_clause = ctx.limitRowCount()
        if limit_clause:
            # TODO: Assign this to the query object
            query.limit = self.visit(limit_clause)
        return query

    @overrides
    def visitSetOperation(
        self, ctx: SqlBaseParser.SetOperationContext
    ) -> SetQuery:
        left_query = self.visit(ctx.left)
        right_query = self.visit(ctx.right)
        kwargs = {}
        if ctx.setQuantifier():
            kwargs["set_quantifier"] = self.visit(ctx.setQuantifier())
        operator_text = ctx.operator.text.upper()
        if operator_text == "INTERSECT":
            return IntersectQuery(left_query, right_query, **kwargs)
        elif operator_text == "UNION":
            return UnionQuery(left_query, right_query, **kwargs)
        elif operator_text == "EXCEPT":
            return ExceptQuery(left_query, right_query, **kwargs)
        raise NotImplementedError(f"Unsupported operator type {operator_text}")

    @overrides
    def visitRowCount(self, ctx: SqlBaseParser.RowCountContext) -> int:
        assert (
            not ctx.QUESTION_MARK()
        ), "Question mark (?) as a row count is currently not supported"
        return int(ctx.INTEGER_VALUE().getText())

    @overrides
    def visitSortItem(self, ctx: SqlBaseParser.SortItemContext) -> OrderTerm:
        value = self.visit(ctx.expression())
        order_type = (
            OrderType.ASC
            if not ctx.ordering or ctx.ordering.text.upper() == "ASC"
            else OrderType.DESC
        )
        null_order = (
            NullOrder.LAST
            if not ctx.nullOrdering or ctx.nullOrdering.text.upper() == "LAST"
            else NullOrder.FIRST
        )
        return OrderTerm(value, order_type, null_order)

    @overrides
    def visitLimitRowCount(
        self, ctx: SqlBaseParser.LimitRowCountContext
    ) -> Optional[int]:
        count = ctx.getText()
        if count.isnumeric():
            return int(count)
        else:
            assert (
                count == "ALL"
            ), "LIMIT quantities can either be numeric or ALL"
            return None

    @overrides
    def visitQueryTermDefault(
        self, ctx: SqlBaseParser.QueryTermDefaultContext
    ) -> SelectQuery:
        return self.visit(ctx.queryPrimary())

    @overrides
    def visitTable(self, ctx: SqlBaseParser.TableContext) -> TableQuery:
        return TableQuery(
            table=table_from_qualifiers(self.visit(ctx.qualifiedName()))
        )

    @overrides
    def visitInlineTable(
        self, ctx: SqlBaseParser.InlineTableContext
    ) -> ValuesQuery:
        return ValuesQuery(
            exprs=[self.visit(expr) for expr in ctx.expression()]
        )

    @overrides
    def visitQueryPrimaryDefault(
        self, ctx: SqlBaseParser.QueryPrimaryDefaultContext
    ) -> Relation:
        query_spec = ctx.querySpecification()
        if not isinstance(query_spec, SqlBaseParser.QuerySpecificationContext):
            raise NotImplementedError(
                "Directly referencing tables without a SELECT clause is not yet implemented"
            )
        return self.visit(query_spec)

    @overrides
    def visitSetQuantifier(
        self, ctx: SqlBaseParser.SetQuantifierContext
    ) -> SetQuantifier:
        return SetQuantifier[ctx.getText()]

    @overrides
    def visitAnd_(self, ctx: SqlBaseParser.And_Context) -> And:
        return And(
            self.visit(ctx.booleanExpression(0)),
            self.visit(ctx.booleanExpression(1)),
        )

    @overrides
    def visitOr_(self, ctx: SqlBaseParser.Or_Context) -> Or:
        return Or(
            self.visit(ctx.booleanExpression(0)),
            self.visit(ctx.booleanExpression(1)),
        )

    @overrides
    def visitLogicalNot(self, ctx: SqlBaseParser.LogicalNotContext) -> Value:
        return ~self.visit(ctx.booleanExpression())

    @overrides
    def visitPredicated(self, ctx: SqlBaseParser.PredicatedContext) -> Value:
        value = self.visit(ctx.valueExpression())
        # If predicate is not passed in, then the boolean expression is really a value expression.
        predicate = ctx.predicate()
        if predicate:
            # HACK: temporarily assign the value as a member to predicate
            predicate.left_value = value
            value = self.visit(predicate)
            predicate.left_value = None
        return value

    @overrides
    def visitComparison(
        self,
        ctx: SqlBaseParser.ComparisonContext,
        left_value: Optional[Value] = None,
    ) -> Value:
        """Visits a comparison between two expressions. However, the grammar here only contains
        the right expression, so we have a left_value member added to the context.
        """
        right_value = self.visit(ctx.right)
        return apply_operator(
            ctx.comparisonOperator().getText(), ctx.left_value, right_value
        )

    @overrides
    def visitQuantifiedComparison(
        self, ctx: SqlBaseParser.QuantifiedComparisonContext
    ) -> None:
        raise NotImplementedError("Quantified comparison is not yet supported")

    @overrides
    def visitInSubquery(self, ctx: SqlBaseParser.InSubqueryContext) -> Value:
        in_query = InQuery(value=ctx.left_value, query=self.visit(ctx.query()))
        if ctx.NOT():
            return ~in_query
        return in_query

    @overrides
    def visitBetween(self, ctx: SqlBaseParser.BetweenContext) -> Value:
        between = Between(
            ctx.left_value,
            lower=self.visit(ctx.lower),
            upper=self.visit(ctx.upper),
        )
        if ctx.NOT():
            return ~between
        return between

    @overrides
    def visitInList(self, ctx: SqlBaseParser.InListContext) -> Value:
        expressions = ctx.expression()
        in_list = InList(
            value=ctx.left_value,
            exprs=[self.visit(expr) for expr in expressions],
        )
        if ctx.NOT():
            return ~in_list
        return in_list

    @overrides
    def visitLike(self, ctx: SqlBaseParser.LikeContext) -> Value:
        escape = None
        if ctx.escape:
            escape = self.visit(ctx.escape)
        like = Like(
            value=ctx.left_value, pattern=self.visit(ctx.pattern), escape=escape
        )
        if ctx.NOT():
            return ~like
        return like

    @overrides
    def visitNullPredicate(
        self, ctx: SqlBaseParser.NullPredicateContext
    ) -> Value:
        is_null = IsNull(ctx.left_value)
        if ctx.NOT():
            return ~is_null
        return is_null

    @overrides
    def visitDistinctFrom(
        self, ctx: SqlBaseParser.DistinctFromContext
    ) -> Value:
        distinct = DistinctFrom(ctx.left_value, right=self.visit(ctx.right))
        if ctx.NOT():
            return ~distinct
        return distinct

    @overrides
    def visitArithmeticBinary(
        self, ctx: SqlBaseParser.ArithmeticBinaryContext
    ) -> Value:
        left = self.visit(ctx.left)
        right = self.visit(ctx.right)
        return apply_operator(ctx.operator.text.upper(), left, right)

    @overrides
    def visitArithmeticUnary(
        self, ctx: SqlBaseParser.ArithmeticUnaryContext
    ) -> Value:
        return apply_operator(
            ctx.operator.text.upper(), self.visit(ctx.valueExpression())
        )

    @overrides
    def visitConcatenation(
        self, ctx: SqlBaseParser.ConcatenationContext
    ) -> Concatenate:
        return Concatenate(values=[self.visit(ctx.left), self.visit(ctx.right)])

    @overrides
    def visitSimpleCase(self, ctx: SqlBaseParser.SimpleCaseContext) -> Case:
        whens = [self.visit(when_) for when_ in ctx.whenClause()]
        value = self.visit(ctx.operand)
        else_ = (
            Else(self.visit(ctx.elseExpression)) if ctx.elseExpression else None
        )
        return Case(branches=whens, else_=else_, value=value)

    @overrides
    def visitSearchedCase(self, ctx: SqlBaseParser.SearchedCaseContext) -> Case:
        whens = [self.visit(when_) for when_ in ctx.whenClause()]
        else_ = (
            Else(self.visit(ctx.elseExpression)) if ctx.elseExpression else None
        )
        return Case(branches=whens, else_=else_)

    @overrides
    def visitWhenClause(self, ctx: SqlBaseParser.WhenClauseContext) -> When:
        return When(
            condition=self.visit(ctx.condition), value=self.visit(ctx.result)
        )

    @overrides
    def visitCast(self, ctx: SqlBaseParser.CastContext) -> Union[Cast, TryCast]:
        expr = self.visit(ctx.expression())
        output_type = self.visit(ctx.type_())
        if ctx.CAST():
            return Cast(expr=expr, data_type=output_type)
        if ctx.TRY_CAST():
            return TryCast(expr=expr, data_type=output_type)
        raise ValueError(f"Unepxected cast expression {ctx.getText()}")

    @overrides
    def visitGenericType(
        self, ctx: SqlBaseParser.GenericTypeContext
    ) -> DataType:
        param_values = [self.visit(param) for param in ctx.typeParameter()]
        type_name = self.visit(ctx.identifier()).upper()
        # We assume the parameters will be passed into here.
        parameters = {
            param.name: val
            for val, param in zip(param_values, FIELDS[type_name])
        }
        return DataType(type_name, parameters=parameters)

    @overrides
    def visitRowType(self, ctx: SqlBaseParser.RowTypeContext) -> DataType:
        types = [self.visit(row) for row in ctx.rowField()]
        return row(dtypes=types)

    @overrides
    def visitRowField(self, ctx: SqlBaseParser.RowFieldContext) -> DataType:
        assert (
            ctx.identifier() is None
        ), "Data types with identifiers currently not supported."
        return self.visit(ctx.type_())

    @overrides
    def visitInterval(self, ctx: SqlBaseParser.IntervalContext) -> Interval:
        string_value = self.visit(ctx.string())
        # Unary plus is essentially a no-op.
        if string_value.startswith("+"):
            string_value = string_value.lstrip("+")

        # There's a pesky sign that can appear outside of the string. We do some inference on the string
        # and push the negation inside the string. I've verified that -"3-100" YEAR TO MONTH is equivalent to "-3-100".
        if ctx.sign and ctx.sign.text == "-":
            # I've tested this and the unary operator +/- can only be applied once to the values inside the string.
            # This means the string can either be +value, -value, or value.
            if string_value.startswith("-"):
                string_value = string_value.lstrip("-")
            else:
                string_value = "-" + string_value
        parameters = {
            "value": string_value,
            "from_interval": self.visit(ctx.from_),
        }
        if ctx.to is not None:
            parameters["to_interval"] = self.visit(ctx.to)
        return Interval(**parameters)

    @overrides
    def visitIntervalField(
        self, ctx: SqlBaseParser.IntervalFieldContext
    ) -> str:
        return ctx.getText()

    @overrides
    def visitIntervalType(
        self, ctx: SqlBaseParser.IntervalTypeContext
    ) -> DataType:
        parameters = {"from_interval": self.visit(ctx.from_)}
        if ctx.to:
            parameters["to_interval"] = self.visit(ctx.to)
        return interval(**parameters)

    @overrides
    def visitArrayType(self, ctx: SqlBaseParser.ArrayTypeContext) -> DataType:
        assert not ctx.INTEGER_VALUE(), "Explicit array size not supported"
        return array(dtype=self.visit(ctx.type_()))

    @overrides
    def visitTypeParameter(
        self, ctx: SqlBaseParser.TypeParameterContext
    ) -> Union[int, DataType]:
        if ctx.type_():
            return self.visit(ctx.type_())
        else:
            assert ctx.INTEGER_VALUE()
            return int(ctx.INTEGER_VALUE().getText())

    @overrides
    def visitDateTimeType(
        self, ctx: SqlBaseParser.DateTimeTypeContext
    ) -> DataType:
        parameters = {}
        data_type = TIMESTAMP if ctx.TIMESTAMP() else TIME

        if ctx.precision:
            parameters["precision"] = self.visit(ctx.precision)
        if ctx.WITH() and ctx.TIME() and ctx.ZONE():
            parameters["timezone"] = True
        else:
            parameters["timezone"] = False
        return DataType(data_type, parameters=parameters)

    @overrides
    def visitDoublePrecisionType(
        self, ctx: SqlBaseParser.DoublePrecisionTypeContext
    ) -> DataType:
        return double()

    @overrides
    def visitLegacyArrayType(
        self, ctx: SqlBaseParser.LegacyArrayTypeContext
    ) -> DataType:
        raise NotImplementedError("Legacy array type not yet supported")

    @overrides
    def visitLegacyMapType(
        self, ctx: SqlBaseParser.LegacyMapTypeContext
    ) -> DataType:
        raise NotImplementedError("Legacy map type not yet supported")

    @overrides
    def visitDereference(self, ctx: SqlBaseParser.DereferenceContext) -> Field:
        primary_expr = self.visit(ctx.primaryExpression())
        if isinstance(primary_expr, Field):
            table = primary_expr.name
            assert (
                not primary_expr.table
            ), "Treeno currently only supports dereference in the form x or y.x"
        else:
            table = primary_expr
        return Field(self.visit(ctx.fieldName), table)

    @overrides
    def visitSpecialDateTimeFunction(
        self, ctx: SqlBaseParser.SpecialDateTimeFunctionContext
    ) -> Function:
        fn_name = ctx.name.text.upper()
        assert (
            fn_name in NAMES_TO_FUNCTIONS
        ), f"Function name {fn_name} not registered in treeno.functions"
        fn = NAMES_TO_FUNCTIONS[fn_name]
        args = []
        if ctx.precision:
            args.append(int(ctx.precision.text))
        return fn(*args)

    @overrides
    def visitCurrentUser(
        self, ctx: SqlBaseParser.CurrentUserContext
    ) -> CurrentUser:
        return CurrentUser()

    @overrides
    def visitCurrentCatalog(
        self, ctx: SqlBaseParser.CurrentCatalogContext
    ) -> CurrentCatalog:
        return CurrentCatalog()

    @overrides
    def visitCurrentSchema(
        self, ctx: SqlBaseParser.CurrentSchemaContext
    ) -> CurrentSchema:
        return CurrentSchema()

    @overrides
    def visitCurrentPath(
        self, ctx: SqlBaseParser.CurrentPathContext
    ) -> CurrentPath:
        return CurrentPath()

    @overrides
    def visitNullLiteral(
        self, ctx: SqlBaseParser.NullLiteralContext
    ) -> Literal:
        # A null literal doesn't have a well-defined type, since it can be any type
        # due to all types in Trino being optional.
        return Literal(None, data_type=unknown())

    @overrides
    def visitNumericLiteral(
        self, ctx: SqlBaseParser.NumericLiteralContext
    ) -> Literal:
        # IMPORTANT: Note that number is only used in primary expression, so it's okay
        # that we get a literal from here. However, if number changes to support other
        # semantic meaning that's not a Literal value, then we need to change this.
        return self.visit(ctx.number())

    @overrides
    def visitIntegerLiteral(
        self, ctx: SqlBaseParser.IntegerLiteralContext
    ) -> Literal:
        value = int(ctx.INTEGER_VALUE().getText())
        if ctx.MINUS() is not None:
            value = -value
        return Literal(value, data_type=infer_integral(value))

    @overrides
    def visitDecimalLiteral(
        self, ctx: SqlBaseParser.DecimalLiteralContext
    ) -> Literal:
        text = ctx.DECIMAL_VALUE().getText()
        negative = ctx.MINUS() is not None

        dtype: DataType
        value: Union[int, Decimal]
        if "." in text:
            value = Decimal(text)
            dtype = infer_decimal(value)
        else:
            value = int(text)
            dtype = integer()

        if negative:
            value = -value

        return Literal(value, data_type=dtype)

    @overrides
    def visitStringLiteral(
        self, ctx: SqlBaseParser.StringLiteralContext
    ) -> Literal:
        string = self.visit(ctx.string())
        return Literal(string, data_type=varchar(max_chars=len(string)))

    @overrides
    def visitDoubleLiteral(
        self, ctx: SqlBaseParser.DoubleLiteralContext
    ) -> Literal:
        number_string = ctx.DOUBLE_VALUE().getText()
        if ctx.MINUS() is not None:
            number_string = ctx.MINUS().getText() + number_string
        return Literal(float(number_string), data_type=double())

    @overrides
    def visitBasicStringLiteral(
        self, ctx: SqlBaseParser.BasicStringLiteralContext
    ) -> str:
        return ctx.getText().strip("'")

    @overrides
    def visitUnicodeStringLiteral(
        self, ctx: SqlBaseParser.UnicodeStringLiteralContext
    ) -> str:
        if ctx.UESCAPE():
            assert ctx.STRING(), "Escape string must be supplied for unicode"
            escape_seq = ctx.STRING().getText().strip("'")
        else:
            escape_seq = "\\"
        s = (
            ctx.UNICODE_STRING()
            .getText()
            .strip("U&")
            .strip("'")
            .replace(escape_seq, "\\u")
            .encode("raw_unicode_escape")
            .decode("unicode_escape")
        )
        return s

    @overrides
    def visitBooleanLiteral(
        self, ctx: SqlBaseParser.BooleanLiteralContext
    ) -> Literal:
        return Literal(self.visit(ctx.booleanValue()), data_type=boolean())

    @overrides
    def visitBooleanValue(self, ctx: SqlBaseParser.BooleanValueContext) -> bool:
        return ctx.TRUE() is not None

    @overrides
    def visitTypeConstructor(
        self, ctx: SqlBaseParser.TypeConstructorContext
    ) -> TypeConstructor:
        value = self.visit(ctx.string())
        if ctx.DOUBLE() and ctx.PRECISION():
            return TypeConstructor(value, data_type=double())
        # It appears the type constructor is fairly primitive in that it doesn't allow parametrized types, like
        # SELECT DECIMAL(30) '3'
        # which means we can assume it's a generic(but inferred) nonparametrized data type.
        type_name = self.visit(ctx.identifier())
        return TypeConstructor(value=value, type_name=type_name)

    @overrides
    def visitFunctionCall(
        self, ctx: SqlBaseParser.FunctionCallContext
    ) -> Function:
        assert (
            not ctx.processingMode()
        ), "Pattern recognition is currently not supported"

        # Qualified names usually have multiple parts, but afaik functions aren't namespaced so there should only
        # be one part
        qual_name = self.visit(ctx.qualifiedName())
        assert (
            len(qual_name) == 1
        ), f"Invalid function name {'.'.join(qual_name)}"

        fn_name = qual_name[0].upper()
        assert (
            fn_name in NAMES_TO_FUNCTIONS
        ), f"Function name {fn_name} not registered in treeno.functions"
        fn = NAMES_TO_FUNCTIONS[fn_name]

        kwargs = {}

        if ctx.over():
            assert issubclass(
                fn, AggregateFunction
            ), "Can't scan over windows on non-aggregate functions"
            kwargs["window"] = self.visit(ctx.over())

        sort_items = ctx.sortItem()
        if sort_items:
            kwargs["orderby"] = [
                self.visit(sort_item) for sort_item in sort_items
            ]

        filter_ = ctx.filter_()
        if filter_:
            kwargs["filter_"] = self.visit(filter_)

        null_treatment = ctx.nullTreatment()
        if null_treatment:
            kwargs["null_treatment"] = self.visit(null_treatment)

        expressions: List[Value]
        if ctx.ASTERISK():
            star: Star
            if ctx.label:
                # If this is an identifier, treat it as an aliased relation
                label = self.visit(ctx.label)
                table = label.name if isinstance(label, Field) else label
                star = Star(table=table)
            else:
                star = Star()
            expressions = [star]
        else:
            # TODO: Are we missing the empty args case?
            expressions = [self.visit(expr) for expr in ctx.expression()]

        return fn.from_args(*expressions, **kwargs)

    @overrides
    def visitLambda_(self, ctx: SqlBaseParser.Lambda_Context) -> Lambda:
        # TODO: We have to implement better tree traversal to get this to work.
        variables = [
            Lambda.Variable(self.visit(identifier))
            for identifier in ctx.identifier()
        ]
        return Lambda.from_generic_expr(variables, self.visit(ctx.expression()))

    @overrides
    def visitNullTreatment(
        self, ctx: SqlBaseParser.NullTreatmentContext
    ) -> NullTreatment:
        if ctx.IGNORE():
            return NullTreatment.IGNORE
        elif ctx.RESPECT():
            return NullTreatment.RESPECT
        else:
            raise ValueError(f"Null treatment {ctx.getText()} not supported")

    @overrides
    def visitFilter_(self, ctx: SqlBaseParser.Filter_Context) -> Value:
        # For now, filter_ is only used inside of a query, solely to get the boolean value.
        return self.visit(ctx.booleanExpression())

    @overrides
    def visitRowConstructor(
        self, ctx: SqlBaseParser.RowConstructorContext
    ) -> RowConstructor:
        values = [self.visit(expr) for expr in ctx.expression()]
        return RowConstructor(values)

    @overrides
    def visitListagg(self, ctx: SqlBaseParser.ListaggContext) -> ListAgg:
        kwargs = {}
        separator = ctx.string()
        if separator:
            kwargs["separator"] = self.visit(separator)

        overflow_behavior = ctx.listAggOverflowBehavior()
        if overflow_behavior:
            kwargs["overflow_filler"] = self.visit(overflow_behavior)

        sort_items = ctx.sortItem()
        if sort_items:
            kwargs["orderby"] = [self.visit(item) for item in sort_items]
        return ListAgg(self.visit(ctx.expression()), **kwargs)

    @overrides
    def visitListAggOverflowBehavior(
        self, ctx: SqlBaseParser.ListAggOverflowBehaviorContext
    ) -> Optional[OverflowFiller]:
        # This is by default None in the parameter. It's only used in LISTAGG.
        if ctx.ERROR():
            return None
        filler = (
            self.visit(ctx.string())
            if ctx.string()
            else OverflowFiller.DEFAULT_FILLER
        )
        return OverflowFiller(
            count_indication=self.visit(ctx.listaggCountIndication()),
            filler=filler,
        )

    @overrides
    def visitListaggCountIndication(
        self, ctx: SqlBaseParser.ListaggCountIndicationContext
    ) -> CountIndication:
        if ctx.WITHOUT():
            return CountIndication.WITHOUT_COUNT
        elif ctx.WITH():
            return CountIndication.WITH_COUNT
        raise ValueError(f"Unknown count indication mode {ctx.getText()}")

    @overrides
    def visitParameter(self, ctx: SqlBaseParser.ParameterContext) -> Literal:
        raise NotImplementedError("Parameters currently not supported")

    @overrides
    def visitOver(self, ctx: SqlBaseParser.OverContext) -> Window:
        """The window can either be an identifier or a full window specification.
        """
        if ctx.windowName:
            return Window(parent_window=self.visit(ctx.windowName))
        return self.visit(ctx.windowSpecification())

    @overrides
    def visitArrayConstructor(
        self, ctx: SqlBaseParser.ArrayConstructorContext
    ) -> Array:
        return Array([self.visit(expr) for expr in ctx.expression()])

    @overrides
    def visitSubscript(self, ctx: SqlBaseParser.SubscriptContext) -> Subscript:
        return Subscript(
            value=self.visit(ctx.value), index=self.visit(ctx.index)
        )

    @overrides
    def visitValueExpressionDefault(
        self, ctx: SqlBaseParser.ValueExpressionDefaultContext
    ) -> Value:
        return self.visit(ctx.primaryExpression())

    @overrides
    def visitSelectSingle(
        self, ctx: SqlBaseParser.SelectSingleContext
    ) -> Value:
        value = self.visit(ctx.expression())
        identifier = ctx.identifier()
        if identifier:
            alias = self.visit(identifier)
            return AliasedValue(value, alias)
        return value

    @overrides
    def visitQuotedIdentifier(
        self, ctx: SqlBaseParser.QuotedIdentifierContext
    ) -> str:
        return ctx.getText().strip('"')

    @overrides
    def visitUnquotedIdentifier(
        self, ctx: SqlBaseParser.UnquotedIdentifierContext
    ) -> str:
        return ctx.getText()

    @overrides
    def visitBackQuotedIdentifier(
        self, ctx: SqlBaseParser.BackQuotedIdentifierContext
    ) -> str:
        raise NotImplementedError("Trino doesn't support backticks AFAIK")

    @overrides
    def visitDigitIdentifier(self, ctx: SqlBaseParser.DigitIdentifierContext):
        raise NotImplementedError(
            "Trino doesn't support digit identifiers AFAIK"
        )

    @overrides
    def visitColumnReference(
        self, ctx: SqlBaseParser.ColumnReferenceContext
    ) -> Field:
        # A column reference can be one of many forms of identifiers
        identifier = ctx.identifier()
        return Field(self.visit(identifier))

    @overrides
    def visitSubqueryExpression(
        self, ctx: SqlBaseParser.SubqueryExpressionContext
    ) -> SelectQuery:
        return self.visit(ctx.query())

    @overrides
    def visitParenthesizedExpression(
        self, ctx: SqlBaseParser.ParenthesizedExpressionContext
    ) -> Value:
        return self.visit(ctx.expression())

    @overrides
    def visitSelectAll(self, ctx: SqlBaseParser.SelectAllContext) -> Value:
        """Visits a `*` or `"table".*` statement. Returns a Star field.
        """
        if ctx.getText() == "*":
            return Star()
        primary_expr = ctx.primaryExpression()
        table: Optional[Union[str, Value]] = None
        if primary_expr:
            expr = self.visit(primary_expr)
            table = expr.name if isinstance(expr, Field) else expr

        # TODO: This is a bit hacky right now, but the table reference is derived from a ColumnReference which is a field
        # here we just interpret a Field to mean a table reference.
        column_aliases = ctx.columnAliases()
        star = Star(table)
        if column_aliases:
            return AliasedStar(star, self.visit(column_aliases))
        return star

    @overrides
    def visitJoinRelation(self, ctx: SqlBaseParser.JoinRelationContext) -> Join:
        left_relation = self.visit(ctx.left)
        # This part of the Trino grammar is really weird - why make it complicated and have `rightRelation` for
        # most join types but make it `right` for cross join and natural joins?
        right_relation = (
            self.visit(ctx.right)
            if ctx.right
            else self.visit(ctx.rightRelation)
        )
        config: JoinConfig
        if ctx.CROSS() and ctx.JOIN():
            config = JoinConfig(JoinType.CROSS)
        elif ctx.NATURAL():
            config = JoinConfig(
                join_type=self.visit(ctx.joinType()), natural=True
            )
        else:
            config = JoinConfig(
                join_type=self.visit(ctx.joinType()),
                criteria=self.visit(ctx.joinCriteria()),
            )
        return Join(left_relation, right_relation, config=config)

    @overrides
    def visitJoinCriteria(
        self, ctx: SqlBaseParser.JoinCriteriaContext
    ) -> JoinCriteria:
        if ctx.USING():
            return JoinUsingCriteria(
                column_names=[
                    self.visit(identifier) for identifier in ctx.identifier()
                ]
            )
        return JoinOnCriteria(constraint=self.visit(ctx.booleanExpression()))

    @overrides
    def visitJoinType(self, ctx: SqlBaseParser.JoinTypeContext) -> JoinType:
        if ctx.LEFT():
            return JoinType.LEFT
        if ctx.RIGHT():
            return JoinType.RIGHT
        if ctx.FULL():
            return JoinType.OUTER
        return JoinType.INNER

    @overrides
    def visitRelationDefault(
        self, ctx: SqlBaseParser.RelationDefaultContext
    ) -> Relation:
        return self.visit(ctx.sampledRelation())

    @overrides
    def visitSampledRelation(
        self, ctx: SqlBaseParser.SampledRelationContext
    ) -> Relation:
        relation = self.visit(ctx.patternRecognition())
        if ctx.TABLESAMPLE():
            return TableSample(
                relation,
                self.visit(ctx.sampleType()),
                self.visit(ctx.percentage),
            )
        return relation

    @overrides
    def visitSampleType(
        self, ctx: SqlBaseParser.SampleTypeContext
    ) -> SampleType:
        return SampleType[ctx.getText()]

    @overrides
    def visitPatternRecognition(
        self, ctx: SqlBaseParser.PatternRecognitionContext
    ) -> Relation:
        relation = self.visit(ctx.aliasedRelation())
        # TODO: Implement MATCH RECOGNIZE
        assert not ctx.MATCH_RECOGNIZE(), "Match recognizes are not supported"
        return relation

    @overrides
    def visitColumnAliases(
        self, ctx: SqlBaseParser.ColumnAliasesContext
    ) -> List[str]:
        return [self.visit(identifier) for identifier in ctx.identifier()]

    @overrides
    def visitQualifiedName(
        self, ctx: SqlBaseParser.QualifiedNameContext
    ) -> List[str]:
        return [self.visit(identifier) for identifier in ctx.identifier()]

    @overrides
    def visitTableName(self, ctx: SqlBaseParser.TableNameContext) -> Table:
        assert not ctx.queryPeriod(), "Query period not supported"
        qualifiers = self.visit(ctx.qualifiedName())
        return table_from_qualifiers(qualifiers)

    @overrides
    def visitSubqueryRelation(
        self, ctx: SqlBaseParser.SubqueryRelationContext
    ) -> SelectQuery:
        return self.visit(ctx.query())

    @overrides
    def visitParenthesizedRelation(
        self, ctx: SqlBaseParser.ParenthesizedRelationContext
    ) -> Relation:
        return self.visit(ctx.relation())

    @overrides
    def visitUnnest(self, ctx: SqlBaseParser.UnnestContext) -> Unnest:
        array_values = [self.visit(expr) for expr in ctx.expression()]
        with_ordinality = ctx.ORDINALITY() is not None
        return Unnest(arrays=array_values, with_ordinality=with_ordinality)

    @overrides
    def visitLateral(self, ctx: SqlBaseParser.LateralContext) -> Lateral:
        return Lateral(subquery=self.visit(ctx.query()))

    @overrides
    def visitAliasedRelation(
        self, ctx: SqlBaseParser.AliasedRelationContext
    ) -> Relation:
        relation = self.visit(ctx.relationPrimary())

        identifier = ctx.identifier()
        if not identifier:
            return relation

        alias = AliasedRelation(relation, self.visit(identifier))
        column_aliases = ctx.columnAliases()
        if column_aliases:
            alias.column_aliases = self.visit(column_aliases)

        return alias

    @overrides
    def visitGroupBy(self, ctx: SqlBaseParser.GroupByContext) -> GroupBy:
        kwargs: Dict[str, Any] = {
            "groups": [self.visit(group) for group in ctx.groupingElement()]
        }
        quantifier = ctx.setQuantifier()
        if quantifier:
            kwargs["groupby_quantifier"] = self.visit(quantifier)
        return GroupBy(**kwargs)

    @overrides
    def visitSingleGroupingSet(
        self, ctx: SqlBaseParser.SingleGroupingSetContext
    ) -> GroupingSet:
        return self.visit(ctx.groupingSet())

    @overrides
    def visitGroupingSet(
        self, ctx: SqlBaseParser.GroupingSetContext
    ) -> GroupingSet:
        return GroupingSet([self.visit(expr) for expr in ctx.expression()])

    @overrides
    def visitRollup(self, ctx: SqlBaseParser.RollupContext) -> Rollup:
        return Rollup([self.visit(expr) for expr in ctx.expression()])

    @overrides
    def visitCube(self, ctx: SqlBaseParser.CubeContext) -> Cube:
        return Cube([self.visit(expr) for expr in ctx.expression()])

    @overrides
    def visitMultipleGroupingSets(
        self, ctx: SqlBaseParser.MultipleGroupingSetsContext
    ) -> GroupingSetList:
        return GroupingSetList(
            groups=[
                self.visit(grouping_set) for grouping_set in ctx.groupingSet()
            ]
        )

    @overrides
    def visitWindowDefinition(
        self, ctx: SqlBaseParser.WindowDefinitionContext
    ) -> Tuple[str, Window]:
        return (self.visit(ctx.name), self.visit(ctx.windowSpecification()))

    @overrides
    def visitWindowSpecification(
        self, ctx: SqlBaseParser.WindowSpecificationContext
    ) -> Window:
        window_frame = ctx.windowFrame()
        # Set to default Window type if not specified
        window = self.visit(window_frame) if window_frame else Window()
        if ctx.ORDER() and ctx.BY():
            window.orderby = [
                self.visit(sort_item) for sort_item in ctx.sortItem()
            ]
        if ctx.partition:
            window.partitions = [
                self.visit(partition) for partition in ctx.partition
            ]
        if ctx.existingWindowName:
            window.parent_window = self.visit(ctx.existingWindowName)
        return window

    @overrides
    def visitWindowFrame(self, ctx: SqlBaseParser.WindowFrameContext) -> Window:
        assert not ctx.measureDefinition(), "MEASURES currently not supported"
        assert not ctx.skipTo(), "AFTER MATCH currently not supported"
        assert (
            not ctx.INITIAL() and not ctx.SEEK()
        ), "INITIAL/SEEK currently not supported"
        assert not ctx.rowPattern(), "Pattern matching currently not supported"
        assert not ctx.subsetDefinition(), "SUBSET currently not supported"
        assert (
            not ctx.variableDefinition()
        ), "Variable definition in window currently not supported"
        window = self.visit(ctx.frameExtent())
        return window

    @overrides
    def visitFrameExtent(self, ctx: SqlBaseParser.FrameExtentContext) -> Window:
        params = {
            "frame_type": FrameType[ctx.frameType.text.upper()],
            "start_bound": self.visit(ctx.start),
        }
        if ctx.end:
            params["end_bound"] = self.visit(ctx.end)
        return Window(**params)

    @overrides
    def visitBoundedFrame(
        self, ctx: SqlBaseParser.BoundedFrameContext
    ) -> BoundedFrameBound:
        return BoundedFrameBound(
            bound_type=BoundType[ctx.boundType.text.upper()],
            offset=self.visit(ctx.expression()),
        )

    @overrides
    def visitUnboundedFrame(
        self, ctx: SqlBaseParser.UnboundedFrameContext
    ) -> UnboundedFrameBound:
        return UnboundedFrameBound(
            bound_type=BoundType[ctx.boundType.text.upper()]
        )

    @overrides
    def visitCurrentRowBound(
        self, ctx: SqlBaseParser.CurrentRowBoundContext
    ) -> CurrentFrameBound:
        return CurrentFrameBound()

    @overrides
    def visitQuerySpecification(
        self, ctx: SqlBaseParser.QuerySpecificationContext
    ) -> SelectQuery:
        # Always returns a list of items to select from
        select_terms = ctx.selectItem()
        query_builder = SelectQuery(
            select=[self.visit(item) for item in select_terms]
        )

        relations = ctx.relation()
        if relations:
            # TODO: These must be CROSS JOIN'ed with each other.
            # currently we don't support JOINs
            if len(relations) > 1:
                raise NotImplementedError(
                    "Currently multiple FROM relations are not supported"
                )
            query_builder.from_ = self.visit(relations[0])

        # Dictates whether we select ALL rows or DISTINCT rows (all by default)
        set_qualifier = ctx.setQuantifier()
        if set_qualifier:
            query_builder.select_quantifier = self.visit(set_qualifier)

        if ctx.where:
            query_builder.where = self.visit(ctx.where)
        if ctx.having:
            query_builder.having = self.visit(ctx.having)
        groupby = ctx.groupBy()
        if groupby:
            query_builder.groupby = self.visit(groupby)
        if ctx.WINDOW():
            query_builder.window = dict(
                self.visit(window_def) for window_def in ctx.windowDefinition()
            )
        return query_builder



"""Coercions between data types, which is important for inferring the ouput type in a typical binary operation like
Add, Subtract, etc in treeno.expressions

The way Trino does it is by finding the correct argument template for a given function and then attempting to coerce
all arguments to the template using TypeCoercion.java. We decide to do it more flexibly during
__attrs_post_init__ to determine the output type, and we perform verifications separately. The conversions here are
not for coercing from a to b, but rather finding the common supertime for a and b.

Note that this module is very much in flux, as I operate under an assumption:
The supertype of the input types under typical binary operators will be the output type (with exceptions, such as
timestamp - timestamp = interval)
"""
from treeno.datatypes.builder import (
    array,
    char,
    decimal,
    hll,
    map_,
    row,
    time,
    timestamp,
    unknown,
    varchar,
)
from treeno.datatypes.types import (
    ARRAY,
    BIGINT,
    CHAR,
    DATE,
    DECIMAL,
    DOUBLE,
    HLL,
    INTEGER,
    INTERVAL,
    MAP,
    P4HLL,
    REAL,
    ROW,
    SMALLINT,
    TIME,
    TIMESTAMP,
    TINYINT,
    VARCHAR,
    DataType,
)

INTEGRAL_TYPES = {INTEGER, BIGINT, SMALLINT, TINYINT}
INTEGRAL_DECIMAL_PRECISION = {TINYINT: 3, SMALLINT: 5, INTEGER: 10, BIGINT: 19}
FLOAT_TYPES = {REAL, DOUBLE}
# Decimal can be either a float or an integral, and we treat it fairly differently so it's a special case
NUMERIC_TYPES = INTEGRAL_TYPES | FLOAT_TYPES | {DECIMAL}

STRING_TYPES = {CHAR, VARCHAR}
# Refer to: https://github.com/trinodb/trino/blob/f382a176760975fec51e89f3dc62060e20420f82/core/trino-spi/src/main/java/io/trino/spi/type/CharType.java#L48
MAX_CHAR_LENGTH = 65536

HLL_TYPES = {HLL, P4HLL}

DATETIME_TYPES = {DATE, TIME, TIMESTAMP}


def get_arithmetic_type(dtype1: DataType, dtype2: DataType) -> DataType:
    if dtype1 == dtype2:
        return dtype1
    name1, name2 = dtype1.type_name, dtype2.type_name
    if name1 == name2:
        return common_same_type(dtype1, dtype2)
    elif {name1, name2}.issubset(NUMERIC_TYPES):
        return common_numeric(dtype1, dtype2)
    return unknown()


def common_supertype(dtype1: DataType, dtype2: DataType) -> DataType:
    """Returns a common supertype for the given dtypes.
    """
    if dtype1.type_name == dtype2.type_name:
        if dtype1 == dtype2:
            return dtype1
        if dtype1.type_name in COMMON_CONVERSION_MAP:
            common_same_type(dtype1, dtype2)
        return unknown()
    elif dtype1.type_name in NUMERIC_TYPES:
        return common_numeric(dtype1, dtype2)
    elif dtype1.type_name in STRING_TYPES:
        return common_string(dtype1, dtype2)
    elif dtype1.type_name in HLL_TYPES:
        return common_hll(dtype1, dtype2)
    elif dtype1.type_name in DATETIME_TYPES:
        return common_datetime(dtype1, dtype2)
    raise TypeError(f"No common supertypes between {dtype1} and {dtype2}")


def common_same_type(dtype1: DataType, dtype2: DataType) -> DataType:
    assert (
        dtype1.type_name == dtype2.type_name
        and dtype1.type_name in COMMON_CONVERSION_MAP
    )
    return COMMON_CONVERSION_MAP[dtype1.type_name](dtype1, dtype2)


def common_datetime(dtype1: DataType, dtype2: DataType) -> DataType:
    # NOTE: It is assumed here that these dtypes don't have the same base type
    assert (
        dtype1.type_name in DATETIME_TYPES
        and dtype2.type_name in DATETIME_TYPES
    ), f"common_numeric must be called with types in {DATETIME_TYPES}"
    # date and time are not convertible to each other
    if {dtype1.type_name, dtype2.type_name} == {DATE, TIME}:
        return unknown()
    # Currently there's either TIMESTAMP, TIME or TIMESTAMP, DATE, both of which get promoted to timestamp
    return dtype1 if dtype1.type_name == TIMESTAMP else dtype2


def common_hll(dtype1: DataType, dtype2: DataType) -> DataType:
    # NOTE: It is assumed here that these dtypes don't have the same base type
    assert (
        dtype1.type_name in HLL_TYPES and dtype2.type_name in HLL_TYPES
    ), f"common_numeric must be called with types in {HLL_TYPES}"
    # No matter what we cast to hll
    return hll()


def common_string(dtype1: DataType, dtype2: DataType) -> DataType:
    # NOTE: It is assumed here that these dtypes don't have the same base type
    assert (
        dtype1.type_name in STRING_TYPES and dtype2.type_name in STRING_TYPES
    ), f"common_numeric must be called with types in {STRING_TYPES}"
    # Refer to https://github.com/trinodb/trino/blob/5450e782ca4b764f17465c1a40b914ad6743bd6c/core/trino-main/src/main/java/io/trino/type/TypeCoercion.java#L501
    # for why CHAR can't be convertible to VARCHAR.
    has_char = CHAR in {dtype1.type_name, dtype2.type_name}
    if has_char:
        dtype1 = (
            promote_varchar_to_char(dtype1)
            if dtype1.type_name != CHAR
            else dtype1
        )
        dtype2 = (
            promote_varchar_to_char(dtype2)
            if dtype2.type_name != CHAR
            else dtype2
        )
        return COMMON_CONVERSION_MAP[CHAR](dtype1, dtype2)
    else:
        return COMMON_CONVERSION_MAP[VARCHAR](dtype1, dtype2)


def common_numeric(dtype1: DataType, dtype2: DataType) -> DataType:
    # NOTE: It is assumed here that these dtypes don't have the same base type
    assert (
        dtype1.type_name in NUMERIC_TYPES and dtype2.type_name in NUMERIC_TYPES
    ), f"common_numeric must be called with types in {NUMERIC_TYPES}"
    if dtype1.type_name in INTEGRAL_TYPES:
        if dtype2.type_name in INTEGRAL_TYPES:
            return common_integral(dtype1, dtype2)
        if dtype2.type_name in FLOAT_TYPES:
            return dtype2
        assert (
            dtype2.type_name == DECIMAL
        ), f"Expected a decimal type for {dtype2}"
        return COMMON_CONVERSION_MAP[DECIMAL](
            promote_integral_to_decimal(dtype1), dtype2
        )
    elif dtype1.type_name in FLOAT_TYPES:
        if dtype2.type_name in FLOAT_TYPES:
            return common_float(dtype1, dtype2)
        # Floats have the highest priority
        return dtype1
    else:
        assert (
            dtype1.type_name == DECIMAL
        ), f"Expected a decimal type for {dtype2}"
        if dtype2.type_name in INTEGRAL_TYPES:
            return COMMON_CONVERSION_MAP[DECIMAL](
                dtype1, promote_integral_to_decimal(dtype2)
            )
        if dtype2.type_name in FLOAT_TYPES:
            return dtype2
    raise TypeError(
        f"Unexpected type conversion {dtype1} and {dtype2} unregistered in {NUMERIC_TYPES}"
    )


def promote_varchar_to_char(varchar_dtype: DataType) -> DataType:
    chars = varchar_dtype.parameters.get("max_chars", None)
    if chars is None:
        return char()
    # VARCHAR can be way bigger than char, so we have to be careful to cap it
    return char(max_chars=min(MAX_CHAR_LENGTH, chars))


def promote_integral_to_decimal(dtype: DataType) -> DataType:
    return decimal(
        precision=INTEGRAL_DECIMAL_PRECISION[dtype.type_name], scale=0
    )


def common_integral(dtype1: DataType, dtype2: DataType) -> DataType:
    # NOTE: It is assumed here that these dtypes don't have the same base type
    assert (
        dtype1.type_name in INTEGRAL_TYPES
        and dtype2.type_name in INTEGRAL_TYPES
    ), f"common_integral must be called with types in {INTEGRAL_TYPES}"
    # Since we're gonna use the precision map, we might as well use it here for priority
    p1, p2 = (
        INTEGRAL_DECIMAL_PRECISION[dtype1.type_name],
        INTEGRAL_DECIMAL_PRECISION[dtype2.type_name],
    )
    return dtype1 if p1 > p2 else dtype2


def common_float(dtype1: DataType, dtype2: DataType) -> DataType:
    assert (
        dtype1.type_name in FLOAT_TYPES and dtype2.type_name in FLOAT_TYPES
    ), f"common_float must be called with types in {FLOAT_TYPES}"
    # TODO: So far we only have DOUBLE and FLOAT, so this is a very simple branch
    return dtype1 if dtype1.type_name == DOUBLE else dtype2


def common_decimal(dtype1: DataType, dtype2: DataType) -> DataType:
    assert (
        dtype1.type_name == dtype2.type_name == DECIMAL
    ), "common_decimal must be called with DECIMAL type only"
    s1, s2 = dtype1.parameters["scale"], dtype2.parameters["scale"]
    p1, p2 = dtype1.parameters["precision"], dtype2.parameters["precision"]
    scale = max([s1, s2])
    return decimal(precision=max([p1 - s1, p2 - s2]) + scale, scale=scale)


def common_varchar(dtype1: DataType, dtype2: DataType) -> DataType:
    assert (
        dtype1.type_name == dtype2.type_name == VARCHAR
    ), "common_varchar must be called with VARCHAR type only"
    chars1 = dtype1.parameters.get("max_chars", None)
    chars2 = dtype2.parameters.get("max_chars", None)

    if chars1 is None or chars2 is None:
        return varchar()

    return varchar(max_chars=max(chars1, chars2))


def common_char(dtype1: DataType, dtype2: DataType) -> DataType:
    assert (
        dtype1.type_name == dtype2.type_name == CHAR
    ), "common_char must be called with CHAR type only"
    chars1 = dtype1.parameters.get("max_chars", None)
    chars2 = dtype2.parameters.get("max_chars", None)

    if chars1 is None or chars2 is None:
        return char()

    return char(max_chars=max(chars1, chars2))


def common_row(dtype1: DataType, dtype2: DataType) -> DataType:
    assert (
        dtype1.type_name == dtype2.type_name == ROW
    ), "common_row must be called with ROW type only"
    dtypes = []
    for t1, t2 in zip(dtype1.parameters["dtypes"], dtype2.parameters["dtypes"]):
        dtypes.append(common_supertype(t1, t2))
    return row(dtypes=dtypes)


def common_map(dtype1: DataType, dtype2: DataType) -> DataType:
    assert (
        dtype1.type_name == dtype2.type_name == MAP
    ), "common_map must be called with MAP type only"
    f1, f2 = dtype1.parameters["from_dtype"], dtype2.parameters["from_dtype"]
    t1, t2 = dtype1.parameters["to_dtype"], dtype2.parameters["to_dtype"]
    return map_(
        from_dtype=common_supertype(f1, f2), to_dtype=common_supertype(t1, t2)
    )


def common_array(dtype1: DataType, dtype2: DataType) -> DataType:
    assert (
        dtype1.type_name == dtype2.type_name == ARRAY
    ), "common_array must be called with ARRAY type only"
    t1, t2 = dtype1.parameters["dtype"], dtype2.parameters["dtype"]
    return array(dtype=common_supertype(t1, t2))


def common_timestamp(dtype1: DataType, dtype2: DataType) -> DataType:
    assert (
        dtype1.type_name == dtype2.type_name == TIMESTAMP
    ), "common_timestamp must be called with TIMESTAMPtype only"
    timezoned = dtype1.parameters["timezone"] or dtype2.parameters["timezone"]
    precision = max(
        dtype1.parameters["precision"], dtype2.parameters["precision"]
    )
    return timestamp(timezoned=timezoned, precision=precision)


def common_time(dtype1: DataType, dtype2: DataType) -> DataType:
    assert (
        dtype1.type_name == dtype2.type_name == TIME
    ), "common_time must be called with TIME type only"
    timezoned = dtype1.parameters["timezone"] or dtype2.parameters["timezone"]
    precision = max(
        dtype1.parameters["precision"], dtype2.parameters["precision"]
    )
    return time(timezoned=timezoned, precision=precision)


def common_interval(dtype1: DataType, dtype2: DataType) -> DataType:
    assert (
        dtype1.type_name == dtype2.type_name == INTERVAL
    ), "common_interval must be called with INTERVAL type only"
    if (
        dtype1.parameters["from_interval"] == dtype2.parameters["from_interval"]
        and dtype1.parameters["to_interval"] == dtype2.parameters["to_interval"]
    ):
        return dtype1
    # No common subtype in general
    return unknown()


COMMON_CONVERSION_MAP = {
    DECIMAL: common_decimal,
    VARCHAR: common_varchar,
    CHAR: common_char,
    ROW: common_row,
    ARRAY: common_array,
    MAP: common_map,
    TIME: common_time,
    TIMESTAMP: common_timestamp,
    INTERVAL: common_interval,
}



import copy
from typing import Any, Dict

import attr

from treeno.base import Sql
from treeno.datatypes.builder import row
from treeno.datatypes.types import DataType
from treeno.expression import Field, Star
from treeno.relation import Query, Schema
from treeno.util import children, is_dictlike, is_listlike


def _make_field(field: Field, schema: Schema) -> Field:
    # If field's table is a whole subquery, it would belong solely
    # to that subquery and not the relation we're dealing with.
    if isinstance(field.table, Query):
        raise NotImplementedError("Subqueries are not yet supported")

    matching_fields = []
    for f in schema.fields:
        if (
            f.name == field.name
            and field.table is None
            or field.table in schema.relation_ids
        ):
            matching_fields.append(f)

    # If the schema didn't include the field, it should have an unknown type
    if not len(matching_fields):
        return copy.deepcopy(field)

    # In the case of there being two identical names, we need there to be only 1 unique field for it.
    if field.table is None:
        assert (
            len(matching_fields) == 1
        ), f"Expected 1 global field matching name {field.name}, got {len(matching_fields)}"
        dtype = matching_fields[0].data_type
    else:
        # Find the right table by filtering 1 more time
        filtered_fields = [
            f for f in matching_fields if f.source.identifier() == field.table
        ]
        assert (
            len(filtered_fields) == 1
        ), f"Two ambiguous fields with same name and table found {schema}"
        dtype = filtered_fields[0].data_type
    return Field(name=field.name, table=field.table, data_type=dtype)


def _make_star(star: Star, schema: Schema) -> Star:
    if isinstance(star.table, Query):
        raise NotImplementedError("Subqueries are not yet supported")
    matching_fields = schema.fields
    if star.table is not None:
        matching_fields = [
            f for f in matching_fields if f.source.identifier() == star.table
        ]
    dtype = row(dtypes=[f.data_type for f in matching_fields])
    return Star(table=star.table, data_type=dtype)


def resolve_fields(node: Any, schema: Schema) -> Any:
    if is_listlike(node):
        return type(node)(resolve_fields(child, schema) for child in node)
    if is_dictlike(node):
        return type(node)(
            (key, resolve_fields(child, schema)) for key, child in node.items()
        )
    # Everything that includes Sql members must be Sql itself, even if it doesn't have a non-throwing sql() function
    # (see JoinConfig for example)
    if not isinstance(node, Sql):
        return copy.deepcopy(node)
    # We don't want to modify data types
    if isinstance(node, DataType):
        return copy.deepcopy(node)
    if isinstance(node, Field):
        return _make_field(node, schema)
    if isinstance(node, Star):
        return _make_star(node, schema)
    changes: Dict[str, Any] = {}
    for k, v in children(node).items():
        changes[k] = resolve_fields(v, schema)
    return attr.evolve(node, **changes)



from collections import defaultdict
from typing import Any, Callable, Dict, List, Optional

import attr

from treeno.base import PrintOptions, Sql

(
    BOOLEAN,
    INTEGER,
    TINYINT,
    SMALLINT,
    BIGINT,
    REAL,
    DOUBLE,
    DECIMAL,
    VARCHAR,
    CHAR,
    VARBINARY,
    JSON,
    DATE,
    TIME,
    TIMESTAMP,
    INTERVAL,
    IP,
    UUID,
    HLL,
    P4HLL,
    QDIGEST,
    TDIGEST,
    ARRAY,
    MAP,
    ROW,
    # NULL's are unknown types because they can be reinterpreted as any type
    # since all types in Trino are optional
    UNKNOWN,
) = ALLOWED_TYPES = [
    "BOOLEAN",
    "INTEGER",
    "TINYINT",
    "SMALLINT",
    "BIGINT",
    "REAL",
    "DOUBLE",
    "DECIMAL",
    "VARCHAR",
    "CHAR",
    "VARBINARY",
    "JSON",
    "DATE",
    "TIME",
    "TIMESTAMP",
    "INTERVAL",
    "IPADDRESS",
    "UUID",
    "HYPERLOGLOG",
    "P4HYPERLOGLOG",
    "QDIGEST",
    "TDIGEST",
    "ARRAY",
    "MAP",
    "ROW",
    "UNKNOWN",
]


@attr.s
class DataType(Sql):
    """Trino data types are used in cast functions and in type verification in the AST.

    This class should almost never be used by the client. Please refer to builder for the available
    builder functions.
    """

    type_name: str = attr.ib()
    parameters: Dict[str, Any] = attr.ib(factory=dict)

    def __attrs_post_init__(self):
        self.type_name = self.type_name.upper()
        assert (
            self.type_name in ALLOWED_TYPES
        ), f"Type {self.type_name} is not a recognized Trino type."
        for field in FIELDS[self.type_name]:
            if not field.required and field.name not in self.parameters:
                # Unfortunately, because default values can be overridden on the session level,
                # we want to be hesitant in imputing missing values with a fixed default value.
                # For example, see hive.timestamp-precision in https://trino.io/docs/current/connector/hive.html
                # Thus, we only do it when there's a default value
                if field.default is None:
                    continue
                self.parameters[field.name] = field.default

            assert (
                field.name in self.parameters
            ), f"{field.name} not specified. Required for {self.type_name}"
            field_value = self.parameters[field.name]
            assert isinstance(field_value, field.type), (
                f"Field {field.name} for type {self.type_name} must "
                f"be of type {field.type.__name__}. "
                f"Got {type(field_value).__name__} instead with value {field_value})"
            )

        all_parameters = set(param.name for param in FIELDS[self.type_name])
        current_parameters = set(self.parameters)
        assert current_parameters.issubset(
            all_parameters
        ), f"Expected a subset of parameters from {all_parameters}, got {current_parameters} instead"
        validator = VALIDATORS.get(self.type_name, None)
        if validator is not None:
            validator(self)

    def sql(self, opts: PrintOptions) -> str:
        if self.type_name in (TIMESTAMP, TIME):
            return emit_timelike(self)
        elif self.type_name == INTERVAL:
            return emit_interval(self)
        elif self.type_name == ROW:
            return emit_row(self)

        if not self.parameters:
            return self.type_name

        type_params = FIELDS[self.type_name]
        values = []
        # We do this to ensure the order of parameters is outputted correctly.
        # Consider the following situation:
        # We pass in parameters={"scale":10} for decimal, and then we add the default precision after during post init.
        # If we were to iterate on the dict's values we'd end up switching precision and scale if we use the order to
        # print the type.
        for param in type_params:
            assert (
                param.name in self.parameters
            ), f"Missing parameter {param.name} for type {self.type_name}"
            values.append(str(self.parameters[param.name]))

        param_string = ",".join(values)
        return f"{self.type_name}({param_string})"


@attr.s
class TypeParameter:
    name: str = attr.ib()
    required: bool = attr.ib()
    type: type = attr.ib()
    validator: Optional[Callable[..., None]] = attr.ib(default=None)
    # NOTE: None is not a valid default value, it represents the absence of a default.
    default: Optional[Any] = attr.ib(default=None)


def emit_interval(interval: DataType) -> str:
    assert interval.type_name == INTERVAL
    from_interval = interval.parameters["from_interval"]
    to_interval = interval.parameters["to_interval"]
    type_string = f"{interval.type_name} {from_interval} TO {to_interval}"
    return type_string


def emit_timelike(timestamp: DataType) -> str:
    assert timestamp.type_name in (TIMESTAMP, TIME)
    type_string = timestamp.type_name
    precision = timestamp.parameters.get("precision", None)
    if precision is not None:
        type_string += f"({precision})"

    timezone = timestamp.parameters.get("timezone", None)
    if timezone:
        type_string += " WITH TIME ZONE"
    return type_string


def emit_row(row: DataType) -> str:
    assert row.type_name == ROW
    subtype_str = ",".join(str(dtype) for dtype in row.parameters["dtypes"])
    return f"{ROW}({subtype_str})"


def validate_timelike(timelike: DataType) -> None:
    precision = timelike.parameters.get("precision", None)
    if precision is not None:
        assert (
            0 <= precision <= 12
        ), f"Precision of {precision} is not supported"


def validate_single_dtype_parameter(row: DataType) -> None:
    data_type = row.parameters["dtype"]
    assert isinstance(data_type, DataType)


def validate_map(row: DataType) -> None:
    from_type = row.parameters["from_dtype"]
    to_type = row.parameters["to_dtype"]
    assert isinstance(from_type, DataType)
    assert isinstance(to_type, DataType)


def validate_row(row: DataType) -> None:
    data_types = row.parameters["dtypes"]
    assert all(
        isinstance(t, DataType) for t in data_types
    ), "dtypes argument must be a list of DataTypes"


def validate_decimal(decimal: DataType) -> None:
    # TODO: Should we validate the scale as well?
    precision = decimal.parameters["precision"]
    assert 0 <= precision <= 38, f"Precision {precision} is not supported"


def validate_nonparametric(nonparametric: DataType) -> None:
    parameters = nonparametric.parameters
    assert not len(
        parameters
    ), f"Nonparametric data type {nonparametric} does not accept parameters, got {parameters}"


def validate_interval(interval: DataType) -> None:
    from_interval = interval.parameters["from_interval"]
    to_interval = interval.parameters["to_interval"]
    assert from_interval in {
        "YEAR",
        "DAY",
    }, f"From interval must be YEAR or DAY, got {from_interval}"
    if from_interval == "YEAR":
        assert (
            to_interval == "MONTH"
        ), f"Currently only YEAR TO MONTH is allowed, not YEAR TO {to_interval}"
    elif from_interval == "DAY":
        assert (
            to_interval == "SECOND"
        ), f"Currently only DAY TO SECOND is allowed, not DAY TO {to_interval}"


FIELDS: Dict[str, List[TypeParameter]] = defaultdict(
    list,
    {
        DECIMAL: [
            # NOTE: According to Trino docs, precision is not optional, but we can still
            # CAST(3.0 AS DECIMAL) which doesn't have a precision specified (but is inferred).
            # The response from this thread:
            # https://trinodb.slack.com/archives/CFLB9AMBN/p1637464297012200
            # ... should dictate this behavior.
            TypeParameter("precision", required=False, type=int, default=38),
            TypeParameter("scale", required=False, type=int, default=0),
        ],
        # VARCHAR's max_chars is not required because it can be unbounded, but we can't supply a default for it
        # because there is no inf representable in integer space. The absence of the parameter shall denote inf.
        VARCHAR: [TypeParameter("max_chars", required=False, type=int)],
        CHAR: [TypeParameter("max_chars", required=False, type=int, default=1)],
        TIME: [
            # Read TIMESTAMP's explanation for why precision is not required but also doesn't have a default.
            TypeParameter("precision", required=False, type=int),
            TypeParameter("timezone", required=False, type=bool, default=False),
        ],
        TIMESTAMP: [
            # Precision is not required, but we also can't supply a reasonable default because
            # you can use connector session settings to tune the default precision of your queries.
            TypeParameter("precision", required=False, type=int),
            TypeParameter("timezone", required=False, type=bool, default=False),
        ],
        ARRAY: [TypeParameter("dtype", required=True, type=DataType)],
        MAP: [
            TypeParameter("from_dtype", required=True, type=DataType),
            TypeParameter("to_dtype", required=True, type=DataType),
        ],
        ROW: [TypeParameter("dtypes", required=True, type=list)],
        INTERVAL: [
            TypeParameter("from_interval", required=True, type=str),
            TypeParameter("to_interval", required=True, type=str),
        ],
        QDIGEST: [TypeParameter("dtype", required=True, type=DataType)],
    },
)

VALIDATORS: Dict[str, Callable[[DataType], None]] = {
    BOOLEAN: validate_nonparametric,
    INTEGER: validate_nonparametric,
    TINYINT: validate_nonparametric,
    SMALLINT: validate_nonparametric,
    BIGINT: validate_nonparametric,
    REAL: validate_nonparametric,
    DOUBLE: validate_nonparametric,
    VARBINARY: validate_nonparametric,
    JSON: validate_nonparametric,
    DATE: validate_nonparametric,
    IP: validate_nonparametric,
    UUID: validate_nonparametric,
    HLL: validate_nonparametric,
    P4HLL: validate_nonparametric,
    TDIGEST: validate_nonparametric,
    UNKNOWN: validate_nonparametric,
    TIME: validate_timelike,
    TIMESTAMP: validate_timelike,
    QDIGEST: validate_single_dtype_parameter,
    ARRAY: validate_single_dtype_parameter,
    MAP: validate_map,
    ROW: validate_row,
    INTERVAL: validate_interval,
    DECIMAL: validate_decimal,
}



import inspect
from abc import ABC
from typing import Any, Dict, List, Type, TypeVar

import attr

from treeno.base import PrintOptions
from treeno.expression import Expression, Value, value_attr, wrap_literal
from treeno.printer import join_stmts

GenericFunction = TypeVar("GenericFunction", bound="Function")

NAMES_TO_FUNCTIONS: Dict[str, Type["Function"]] = {}
FUNCTIONS_TO_NAMES: Dict[Type["Function"], str] = {}


@value_attr
class Function(Expression, ABC):
    """Functions are expressions that require parenthesizing and support aggregation, filter, sortItem,
    pattern recognition, etc.

    Functions are different from general expressions such as Cast, TypeConstructor, Like, etc.
    """

    def __init_subclass__(cls, *args, **kwargs):
        super().__init_subclass__(*args, **kwargs)

        # TODO: We ignore abstract classes, but some already have all of their
        # abstract methods defined. Thus we do the extra check of ABC as a direct base.
        # See: https://stackoverflow.com/questions/62352982/python-determine-if-class-is-abstract-abc-without-abstractmethod
        if inspect.isabstract(cls) or ABC in cls.__bases__:
            return

        if not hasattr(cls, "FN_NAME"):
            raise TypeError(
                f"Every Function that's not an ABC must have the field FN_NAME defined. {cls.__name__} currently violates this constraint."
            )
        fn_name = getattr(cls, "FN_NAME")
        assert isinstance(fn_name, str), "FN_NAME must be a string"
        NAMES_TO_FUNCTIONS[fn_name] = cls
        FUNCTIONS_TO_NAMES[cls] = fn_name

    @classmethod
    def from_args(
        cls: Type[GenericFunction], *values: Any, **kwargs: Any
    ) -> GenericFunction:
        """This positional-only args constructor is required because there is no easy way to perform function overloading
        in python, so each Function class that has strange inputs should be constructed this way (and have this function
        overridden to do the right thing).
        """
        arguments = attr.fields(cls)
        keyword_only_arguments = set()
        for attribute in arguments:
            if attribute.kw_only and attribute.init:
                keyword_only_arguments.add(attribute.name)
        assert set(kwargs.keys()).issubset(
            keyword_only_arguments
        ), f"No keyword arguments allowed except for {keyword_only_arguments}"
        return cls(*values, **kwargs)

    def to_string(self, values: List[Value], opts: PrintOptions) -> str:
        arg_string = join_stmts([value.sql(opts) for value in values], opts)
        return f"{FUNCTIONS_TO_NAMES[type(self)]}({arg_string})"


@value_attr
class UnaryFunction(Function, ABC):
    value: Value = attr.ib(converter=wrap_literal)

    def sql(self, opts: PrintOptions) -> str:
        return self.to_string([self.value], opts)



import functools
from typing import Any, ClassVar, List, Optional, Type

import attr

from treeno.base import PrintOptions
from treeno.datatypes.builder import unknown
from treeno.datatypes.conversions import common_supertype
from treeno.expression import Value, value_attr, wrap_literal, wrap_literal_list
from treeno.functions.base import Function, GenericFunction


@value_attr
class If(Function):
    FN_NAME: ClassVar[str] = "IF"
    condition: Value = attr.ib(converter=wrap_literal)
    true_value: Value = attr.ib(converter=wrap_literal)
    false_value: Optional[Value] = attr.ib(
        default=None, converter=attr.converters.optional(wrap_literal)
    )

    def __attrs_post_init__(self) -> None:
        if self.false_value is not None:
            self.data_type = common_supertype(
                self.true_value.data_type, self.false_value.data_type
            )
        else:
            self.data_type = self.true_value.data_type

    def sql(self, opts: PrintOptions) -> str:
        args = [self.condition, self.true_value]
        if self.false_value is not None:
            args.append(self.false_value)
        return self.to_string(args, opts)


@value_attr
class Coalesce(Function):
    FN_NAME: ClassVar[str] = "COALESCE"
    values: List[Value] = attr.ib(converter=wrap_literal_list)

    def __attrs_post_init__(self) -> None:
        dtypes = [
            val.data_type for val in self.values if val.data_type != unknown()
        ]
        self.data_type = functools.reduce(
            lambda t1, t2: common_supertype(t1, t2), dtypes
        )

    @classmethod
    def from_args(
        cls: Type[GenericFunction], *args: Any, **kwargs: Any
    ) -> GenericFunction:
        assert not kwargs, "Keyword arguments not allowed for COALESCE"
        return cls(values=args)

    def sql(self, opts: PrintOptions) -> str:
        return self.to_string(self.values, opts)


@value_attr
class NullIf(Function):
    FN_NAME: ClassVar[str] = "NULLIF"
    value1: Value = attr.ib(converter=wrap_literal)
    value2: Value = attr.ib(converter=wrap_literal)

    def __attrs_post_init__(self) -> None:
        self.data_type = self.value1.data_type

    def sql(self, opts: PrintOptions) -> str:
        return self.to_string([self.value1, self.value2], opts)


@value_attr
class Try(Function):
    FN_NAME: ClassVar[str] = "TRY"
    expr: Value = attr.ib(converter=wrap_literal)

    def __attrs_post_init__(self) -> None:
        self.data_type = self.expr.data_type

    def sql(self, opts: PrintOptions) -> str:
        return self.to_string([self.expr], opts)



from typing import ClassVar

import attr

from treeno.base import PrintOptions
from treeno.datatypes.builder import double
from treeno.expression import Value, value_attr, wrap_literal
from treeno.functions.base import Function


@value_attr
class Power(Function):
    FN_NAME: ClassVar[str] = "POWER"
    x: Value = attr.ib(converter=wrap_literal)
    p: Value = attr.ib(converter=wrap_literal)

    def __attrs_post_init__(self) -> None:
        self.data_type = double()

    def sql(self, opts: PrintOptions) -> str:
        return self.to_string([self.x, self.p], opts)



"""Builders for common types as syntactic sugar.

Note that this module is mainly for client use. This module should not be used within treeno.datatypes.types to avoid
circular dependencies.

Note that we explicitly disallow arguments for types that can't be parametrized. We also strictly only allow keyword
only arguments.

TODO: We should replace the kwargs with actual argument names
"""
from treeno.datatypes import types


def boolean() -> types.DataType:
    return types.DataType(types.BOOLEAN)


def tinyint() -> types.DataType:
    return types.DataType(types.TINYINT)


def smallint() -> types.DataType:
    return types.DataType(types.SMALLINT)


def integer() -> types.DataType:
    return types.DataType(types.INTEGER)


def bigint() -> types.DataType:
    return types.DataType(types.BIGINT)


def real() -> types.DataType:
    return types.DataType(types.REAL)


def double() -> types.DataType:
    return types.DataType(types.DOUBLE)


def decimal(**kwargs) -> types.DataType:
    return types.DataType(types.DECIMAL, parameters=kwargs)


def varchar(**kwargs) -> types.DataType:
    return types.DataType(types.VARCHAR, parameters=kwargs)


def char(**kwargs) -> types.DataType:
    return types.DataType(types.CHAR, parameters=kwargs)


def varbinary() -> types.DataType:
    return types.DataType(types.VARBINARY)


def json() -> types.DataType:
    return types.DataType(types.JSON)


def date() -> types.DataType:
    return types.DataType(types.DATE)


def time(**kwargs) -> types.DataType:
    return types.DataType(types.TIME, parameters=kwargs)


def timestamp(**kwargs) -> types.DataType:
    return types.DataType(types.TIMESTAMP, parameters=kwargs)


def interval(**kwargs) -> types.DataType:
    return types.DataType(types.INTERVAL, parameters=kwargs)


def array(**kwargs) -> types.DataType:
    return types.DataType(types.ARRAY, parameters=kwargs)


def map_(**kwargs) -> types.DataType:
    return types.DataType(types.MAP, parameters=kwargs)


def row(**kwargs) -> types.DataType:
    return types.DataType(types.ROW, parameters=kwargs)


def unknown() -> types.DataType:
    return types.DataType(types.UNKNOWN)


def ip() -> types.DataType:
    return types.DataType(types.IP)


def uuid() -> types.DataType:
    return types.DataType(types.UUID)


def hll() -> types.DataType:
    return types.DataType(types.HLL)


def p4hll() -> types.DataType:
    return types.DataType(types.P4HLL)


def qdigest(**kwargs) -> types.DataType:
    return types.DataType(types.QDIGEST, parameters=kwargs)


def tdigest() -> types.DataType:
    return types.DataType(types.TDIGEST)
